/*
 * Copyright 2011 by Kappich Systemberatung, Aachen
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.CommunicationError;
import de.bsvrz.dav.daf.main.ConnectionException;
import de.bsvrz.dav.daf.main.InconsistentLoginException;
import de.bsvrz.dav.daf.main.MissingParameterException;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.client.console.Console;
import de.bsvrz.pua.prot.exceptions.ScopeException;
import de.bsvrz.pua.prot.manager.Manager;
import de.bsvrz.pua.prot.util.MemberCheck;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.DAVAppBase;
import de.bsvrz.sys.funclib.losb.datk.PidProtocol;
import de.bsvrz.sys.funclib.losb.datk.PidScript;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

import java.io.File;

/**
 * Zum Starten der SWE Protokolle und Auswertungen muss diese Klasse verwendet werden. Unterstütze Parameter (neben den üblichen DAV Parametern):<br>
 * <code>konfigurationsVerantwortlicher=<konfigurationsobjekt></code><br> &nbsp;&nbsp;&nbsp;Legt die Pid oder Id des Konfigurationsverantwortlichen fest, der
 * für Archivanfragen und Anfragen and die SWE Protokollerstellung verwendet wird. Default: Lokaler Konfigurationsverantwortlicher.<br>
 * <code>reserveSpeichermedium=[Zahl]</code><br> &nbsp;&nbsp;&nbsp;Speicherplatz in MB der beim Abspeichern von Protokollen frei bleiben muss. Standardmäßig
 * wird der Wert 5 verwendet.<br> <code>skriptVerzeichnis=[Pfad]</code><br> &nbsp;&nbsp;&nbsp;Verzeichnis, in dem Skriptdateien und Protokolle gehalten
 * werden. Das Verzeichnis wird angelegt, falls es noch nicht existiert. Default: ./script<br> <code>-maxProtokolle=[Zahl]</code><br> &nbsp;&nbsp;&nbsp;Maximale
 * Anzahl der parallel zu erstellenden Protokolle. Werden bereits genau so viele Protokolle erstellt wie über diesen Parameter spezifiziert, dann werden alle
 * weiteren Anfragen zur Protokollerstellung abgelehnt. Default: 5.<br> <code>-groesseWarteschlange=[Zahl]</code><br> Legt die maximale Größe der Warteschlange,
 * in der eingehende Aufträge zur Protokollerstellung oder zum Laden gespeicherter Protokolle abgelegt werden, fest. Default: 20.<br>
 * <code>-maxWartezeit=[Sekunden]</code><br> Maximale Zeitdauer in Sekunden, für die die Protokollerstellung vom Client angehalten werden kann. Default: 60
 * Sekunden.<br> <code>-noPrettyPrinter</code><br> &nbsp;&nbsp;&nbsp;Wird dieser Kommandozeilenparameter angegeben, so wird der Pretty Printer nicht verwendet.
 * <code>-kommandoPort=[Zahl]</code><br> &nbsp;&nbsp;&nbsp;Port des Telnet-Servers zum Herunterfahren der Applikation. Default: 3577<br>
 * <code>-kommandoTimeout=[Zahl]</code><br> &nbsp;&nbsp;&nbsp;Timeout des Telnet Servers bei inaktiver Verbindung in Minuten. Default: {@link
 * #TELNET_TIMEOUT}<br> Beispiel:<br> <code> java de.bsvrz.pua.prot.Protocols -konfigurationsVerantwortlicher=kv.testKonfiguration -reserveSpeichermedium=111
 * -skriptVerzeichnis=d:\SkripteUndProtokolle -maxProtokolle=2 -kommandoPort=3577 -debugLevelStdErrText=ALL -benutzer=Tester -authentifizierung=..\..\passwd
 * </code>
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
@SuppressWarnings("deprecation")
public class Protocols extends DAVAppBase {

	/** Debug Ausgaben */
	private Debug _debug = Debug.getLogger();

	/** Name des Lockfiles, um gleichzeitiges Starten von mehreren Pua-instanzen in einem Arbeitsverzeichnis zu verhindern. */
	public static final String LOCKFILE = "_isActive.flag";

	/** Wird zur Kennzeichnung des Defautltwertes verwendet. */
	private static final String DEFAULT_CONFIGURATIONAUTH = "pua.default.value.localconfigurationauthority";

	/** Port der auf Anfragen wartet. */
	public static int TELNET_PORT = 3577;

	/** Timeout des Telnet-Servers. Eine inaktive Verbindung eines Clients zum Telnet-Server wird nach <code>TELNET_TIMEOUT</code> Minuten beendet. */
	public static int TELNET_TIMEOUT = 5;

	/** Optionaler Kommandozeilenparameter für Konfigurationsverantwortlichen. Standardmäßig wird der lokale Konfigurationsverantwortliche verwendet. */
	public static final String CONFIG_AUTH = "-konfigurationsVerantwortlicher=" + DEFAULT_CONFIGURATIONAUTH;

	/** Optionaler Kommandozeilenparameter Speicherplatzbegrenzung. Defaultwert ist 0. */
	public static final String QUOTA = "-reserveSpeichermedium=0";

	/** Standard Pfad der SWE */
	private static final String LASTPATH = "script";

	/** Optionaler Kommandozeilenparameter Skriptverzeichnis. Defaultwert ist ./script . */
	public static final String PATH = "-skriptVerzeichnis=" + LASTPATH;

	/** Optionaler Kommandozeilenparameter Anzahl gleichzeitig erstellbarer Protokolle. Defaultwert ist 5. */
	public static final String MAX_PROTOCOLS = "-maxProtokolle=5";

	/** Optionaler Kommandozeilenparameter Archivanfrage mit Pid (default: ja) */
	public static final String FETCH_OLD_OBJECTS = "-archivAbfrageMitPid=ja";

	/** Optionaler Kommandozeilenparameter Port des Telnet Servers. Standardmäßig wird der Telnet Server auf port {@link #TELNET_PORT} gestartet. */
	public static final String TELNET_SERVER = "-kommandoPort=" + TELNET_PORT;

	/** Timeout des Telnet-Servers. */
	public static final String TELNET_TIMEOUTP = "-kommandoTimeout=" + TELNET_TIMEOUT;

	/** Optionaler Kommandozeilenparameter Pretty Printer verwenden. Standardmäßig wird der Pretty Printer verwendet */
	public static final String NO_PRETTY_PRINTER = "-noPrettyPrinter";

	/**
	 * Optionaler Kommandozeilenparameter Wartenschlangengröße. Legt die maximale Größe der Warteschlange, in der eingehende Aufträge zur Protokollerstellung oder
	 * zum Laden gespeicherter Protokolle abgelegt werden, fest.
	 */
	private static final String MAX_QUERIES = "-groesseWarteschlange=20";

	/** Maximale Zeitdauer in Sekunden, für die die Protokollerstellung vom Client angehalten werden kann. */
	private static final String MAX_STOP_TIME = "-maxWartezeit=60";

	/** Kommandozeilenparamter anzeigen. */
	private static final String HELP1 = "-?";

	/** Kommandozeilenparamter anzeigen. */
	private static final String HELP2 = "/?";

	/** Kommandozeilenparamter anzeigen. */
	private static final String HELP3 = "-help";


	private String _configAuth;

	private int _quota;

	private String _path;

	private int _maxProtocols;

	private int _maxQueries;

	private long _maxDelay;

	// static ist eigentlich unschön, aber da es nur ein Workaround ist, ist es besser, als diesen Parameter
	// durch alle möglichen Klassen und Queues bis zu ArchiveUser durchzureichen.
	public static boolean _fetchOldObjects = true;

	/** Verwalten eingehende Aufträge */
	public Manager _manager;

	/** Gibt an ob der Pretty Printer verwendet werden soll. */
	private boolean _prettyPrinter = true;

	/** Zeigt an ob die SWE heruntergefahren wird. */
	private boolean _shutdown = false;

	/** Monitor, der zusammen mit <code>shutdown</code> benutzt wird, um sicherzustellen, dass der Code zum Beenden von PuA nur einmal ausgeführt wird. */  
	private Object _shutdownLock = new Object();

	/** Flag, das gesetzt ist, wenn die Verbindung zum Datenverteiler aufgebaut wurde und gelöscht wird, wenn die Verbindung zum Datenverteiler unterbrochen wurde.*/
	private volatile boolean _davConnected = false;
	
	/** Anzahl Minuten ohne Antwort vom Client, nach der die Übertragung getrennt wird */
	private int _timeout = 15;


	/**
	 * Initialisierung. Wertet Aufrufparameter aus.
	 *
	 * @param args Kommandozeilenparameter.
	 *
	 * @throws Exception Fehler.
	 * @see DAVAppBase#DAVAppBase(String[],String,boolean)
	 */
	public Protocols(String[] args) throws Exception {
		super(args, "Protokolle und Auswertungen", new DisconnectHandler() {
			// Initialer DisconnectHandler wird später in der run()-Methode ersetzt
			public void handleDisconnect() throws MissingParameterException, InconsistentLoginException, CommunicationError, ConnectionException {
				Debug.getLogger().error("Verbindung zum Datenverteiler wurde frühzeitig terminiert, 'Protokolle und Auswerungen' wird beendet");
				System.exit(1);
			}
		});

		boolean ok = false;
		try {
			_configAuth = argList.fetchArgument(CONFIG_AUTH).asNonEmptyString();
			_quota = argList.fetchArgument(QUOTA).intValueBetween(0, Integer.MAX_VALUE);
			String tempPath = argList.fetchArgument(PATH).asString();
			if(tempPath == null || tempPath.trim().equals("")) {
				this._path = LASTPATH;
			}
			else {
				this._path = tempPath;
			}
			_maxProtocols = argList.fetchArgument(MAX_PROTOCOLS).intValueBetween(1, Integer.MAX_VALUE);
			TELNET_PORT = argList.fetchArgument(TELNET_SERVER).intValueBetween(0, Integer.MAX_VALUE);
			TELNET_TIMEOUT = argList.fetchArgument(TELNET_TIMEOUTP).intValueBetween(1, Integer.MAX_VALUE);
			_prettyPrinter = !(argList.fetchArgument(NO_PRETTY_PRINTER + "=nein").booleanValue());
			_maxQueries = argList.fetchArgument(MAX_QUERIES).intValueBetween(0, Integer.MAX_VALUE);
			_maxDelay = argList.fetchArgument(MAX_STOP_TIME).intValueBetween(0, Integer.MAX_VALUE) * 1000;
			_fetchOldObjects = argList.fetchArgument(FETCH_OLD_OBJECTS).booleanValue();
			_timeout = argList.fetchArgument("-clientTimeout=15").intValueBetween(-1, Integer.MAX_VALUE);

			ok = !(argList.hasArgument(HELP1) || argList.hasArgument(HELP2) || argList.hasArgument(HELP3));
		}
		catch(NumberFormatException e) {
			System.out.println("Zahl erwartet: " + e.getMessage());
		}
		catch(IllegalArgumentException e) {
			System.out.println("Fehlerhafter Aufrufparameter: " + e.getMessage());
		}
		catch(IllegalStateException e) {
			System.out.println("Dem Argument wurde kein Wert zugeordnet: " + e.getMessage());
		}

		try {
			argList.ensureAllArgumentsUsed();
		}
		catch(IllegalStateException e) {
			System.out.println("Unbenutze Argumente: " + e.getMessage());
		}

		if(!ok) {
			System.out.println("Starten von Protokollen und Auswertungen:");
			System.out.println("-----------------------------------------");
			System.out.println("PuA benötigt die folgenden Bibliotheken:");
			System.out.println("antlr_2.7.7.jar, de.bsvrz.sys.funclib.losb, de.bsvrz.sys.funclib.dynobj und die Bibliotheken der Kernsoftware");
			System.out.println("\nKommandozeilenparameter:");
			System.out.println("\tkonfigurationsVerantwortlicher=<konfigurationsobjekt>");
			System.out.println(
					"\t\tLegt die Pid oder Id des Konfigurationsverantwortlichen fest," + "\n\t\tder für Archivanfragen und Anfragen and die SWE"
					+ "\n\t\tProtokollerstellung verwendet wird. Default: Lokaler KV."
			);
			System.out.println("\treserveSpeichermedium=<Zahl>");
			System.out.println(
					"\t\tSpeicherplatz in MB der beim Abspeichern von Protokollen frei" + "\n\t\tbleiben muss. Default: 5 MB"
			);
			System.out.println("\tskriptVerzeichnis=<Pfad>");
			System.out.println(
					"\t\tVerzeichnis, in dem Skriptdateien und Protokolle gehalten werden." + "\n\t\tDefault: Aktuelles Verzeichnis."
			);
			System.out.println("\t-maxProtokolle=<Zahl>");
			System.out.println(
					"\t\tMaximale Anzahl der parallel zu erstellenden Protokolle. Werden" + "\n\t\tbereits genau so viele Protokolle erstellt wie über diesen"
					+ "\n\t\tParameter spezifiziert, dann werden alle weiteren Anfragen" + "\n\t\tzur Protokollerstellung abgelehnt. Default: 5."
			);
			System.out.println("\t-archivAbfrageMitPid=nein");
			System.out.println("\t\tVerhindert, dass bei Archivanfragen historische, gelöschte Objekte mit gleicher Pid mit abgefragt werden.");
			System.out.println("\t-kommandoPort=<Zahl>");
			System.out.println("\t\tPort auf dem der Telnet-Server gestartet wird. Default: 3577.");
			System.out.println("\t-kommandoTimeout=<Zahl>");
			System.out.println("\t\tTimout des Telnet Servers bei inaktiver Verbindung. Default: 5.");
			System.out.println("\t-groesseWarteschlange=<Zahl>");
			System.out.println(
					"\t\tLegt die maximale Größe der Warteschlange, in der eingehende" + "\n\t\tAufträge zur Protokollerstellung oder zum Laden gespeicherter"
					+ "\n\t\tProtokolle abgelegt werden, fest. Default: 20."
			);
			System.out.println("\t-maxWarteZeit=<Sekunden>");
			System.out.println(
					"\t\tMaximale Zeitdauer in Sekunden, für die die Protokollerstellung" + "\n\t\tvom Client angehalten werden kann.Default: 60 Sekunden."
			);	
			System.out.println("\t-clientTimeout=<Minuten>");
			System.out.println(
					"\t\tTimeout in Minuten. Ein Client, der beim Abruf eines Protokolls länger als x Minuten lang keine weiteren Protokolldaten" +
							"\n\t\tmehr anfordert gilt als nicht mehr verbunden und die Datenübertragung wird abgebrochen. Default: 15."
			);
			System.out.println(
					"\nBeispiel:\n\t-konfigurationsVerantwortlicher=kv.testKonfiguration -reserveSpeichermedium=111 -skriptVerzeichnis=d:\\ProtokolleUndSkripte -maxProtokolle=2 -kommandoPort=3577 -debugLevelStdErrText=ALL -benutzer=Tester -authentifizierung=..\\..\\passwd"
			);

			exit("Protokolle und Auswertungen wird beendet.", true);
		}
		else {
			//Verzeichnis ggf. anlegen.
			File puaDir = new File(_path);
			if(!puaDir.exists()) {
				if(puaDir.mkdir()) {
					_debug.info("Verzeichnis für Skripte angelegt: " + puaDir.getAbsolutePath());
				}
				else {
					_debug.error("Verzeichnis für Skripte konnte nicht angelegt werden. " + puaDir.getAbsolutePath());
					System.exit(-1);
				}
			}
			else if(puaDir.isDirectory()) {
				_debug.info("Verwende bereits bestehendes Verzeichnis: " + puaDir.getAbsolutePath());
			}
			else {
				_debug.error("Das angegebene Verzeichnis existiert nicht. Es wurde ein Dateiname angegeben: " + puaDir.getAbsolutePath());
				System.exit(-1);
			}

			//sicherstellen, das keine andere Pua Instanz im gleichen Verzeichnis arbeitet:
			if(!_path.endsWith(File.separator)) _path = _path + File.separator;
			File lockFile = new File(_path + LOCKFILE);
			if(lockFile.exists()) {
				_debug.error(
						"Es wurde bereits eine Instanz der SWE 'Protokolle und Auswertungen' im Verzeichnis " + _path + " gestartet."
						+ "Bitte verwenden Sie ein anderes Arbeitsverzeichnis. Fall Sie sicher sind, dass das Verzeichnis nicht "
						+ "mehr von 'Protokolle und Auswertungen' benutzt wird, dann löschen sie die Datei " + lockFile + " und starten sie die SWE erneut."
				);
				System.exit(-1);
			}
			else {
				if(!lockFile.createNewFile()) {
					_debug.warning("Kann Lock-File " + lockFile + " nicht anlegen.");
				}
				else {
					_debug.finest("Lock-File (" + lockFile + ") angelegt.");
				}
			}
		}
	}


	/**
	 * Zum Starten der WE
	 *
	 * @param args Aufrufparameter (siehe oben)
	 *
	 * @throws Exception Fehler
	 */
	public static void main(String args[]) throws Exception {
		Protocols protocols = new Protocols(args);
		protocols.run();
	}

	/**
	 * Startet Protokolle und Auswertungen. Baut Verbundung zum Dav auf. Startet die Telent-Konsole auf Port {@link #TELNET_PORT} Überpürft ob mit einer
	 * kompatiblen Konfiguration gearbeitet wird.
	 */
	public void run() {
		try {
			super.connectToDav();
			ClientDavInterface dav = super.getDavCon();
			_davConnected = (dav != null);

			setDisconnectHandler(new DisconnectHandler() {
				public void handleDisconnect() throws MissingParameterException, InconsistentLoginException, CommunicationError, ConnectionException {
					_davConnected = false;
					exit("Verbindung zum Datenverteiler wurde terminiert, 'Protokolle und Auswertungen' wird kontrolliert beendet", true);
				}
			});

			Runtime.getRuntime().addShutdownHook(new Thread(new Runnable() {
				public void run() {
					exit("Terminierungssignal empfangen, 'Protokolle und Auswertungen' wird kontrolliert beendet", false);
				}
			}));

			//Prüfen ob kv gültig (Zugriff auf Archiv, Menge der Skriptobjekte, Protokoll-Anfrageschnittstelle
			SystemObject temp;
			if(DEFAULT_CONFIGURATIONAUTH.equals(_configAuth)) {
				temp = dav.getLocalConfigurationAuthority();
			}
			else {
				temp = dav.getDataModel().getObject(_configAuth);
			}
			if(temp == null) throw new FailureException("Ungültige PID des Konfigurationsverantwortlichen" + _configAuth, LoggerException.ERROR);
			if(!(temp instanceof ConfigurationObject)) {
				throw new FailureException(
						_configAuth + " ist kein Konfigurationsobjekt vom Typ ConfigurationObject, sondern ein " + temp.getType(), LoggerException.ERROR
				);
			}

			ConfigurationObject auth = (ConfigurationObject)temp;

			//prüfen ob das Objekt eine Menge für die Skriptobjekte bereitstellt
			if(auth.getMutableSet(PidScript.objectSet) == null) {
				throw new FailureException(_configAuth + " enthält keine Menge " + PidScript.objectSet + " für die Skriptobjekte.", LoggerException.ERROR);
			}
			//Protokoll-Anfrageschnittstelle
			AttributeGroup atg = dav.getDataModel().getAttributeGroup(PidProtocol.atgProtocolRequest);
			MemberCheck.isAttributeGroupMember(auth, atg);

			//Submodule & Telnet-Konsole starten:
			Console console = new Console();
			_manager = new Manager(dav, auth, _path, _maxProtocols, _quota, _maxDelay, _timeout, _maxQueries, console);
			console.createConsole(dav, this, TELNET_PORT, TELNET_TIMEOUT);
			_manager.launch(_prettyPrinter);
			waitTillDone();
		}
		catch(ConfigurationException e) {
			exit(ErrorMessage.COMMUNICATION + e.getMessage(), true);
		}
		catch(FailureException e) {
			exit(e.getMessage(), true);
		}
		catch(ScopeException e) {
			exit(_configAuth + " benötigt die Attributgruppe " + PidProtocol.atgProtocolRequest, true);
		}
		catch(Exception e) {
			exit("Fehler beim Verbinden mit dem Datenverteiler: " + e.getMessage(), true);
		}
	}


	/** Wartet bis die SWE hochgefahren ist. XXX Aktives Warten -> ändern! */
	private void waitTillDone() {
		try {
			while(!isReady() && !_shutdown) Thread.sleep(150);
			if(!_shutdown) _debug.info("Die SWE 'Protokolle und Auswertungen' ist bereit.");
		}
		catch(InterruptedException e) {
			exit(e.getMessage(), true);
		}
	}


	/**
	 * Beendet die Applikation und gibt ggf. eine Fehlermeldung aus.
	 *
	 * @param error      Fehlermeldung. Kann <code>null</code> sein.
	 * @param systemExit <code>true</code> falls die Applikation mit {@link System#exit(int)} beendet werden soll.
	 */
	public void exit(String error, boolean systemExit) {
		try {
			exit(error);
		}
		catch(Exception e) {
			_debug.error("Fehler beim Beenden von 'Protokolle und Auswertungen'", e);
		}
		finally {
			if(systemExit) System.exit(-1);
		}
	}

	/**
	 * Beendet die Applikation und gibt ggf. eine Fehlermeldung aus.
	 *
	 * @param error Fehlermeldung. Kann <code>null</code> sein.
	 */
	public void exit(String error) {
		synchronized(_shutdownLock) {
			if(_shutdown) return;
			_shutdown = true;
		}
		if(error != null && !error.equals("")) {
			_debug.error(error);
			System.out.println(error);
		}
		if(_manager != null) {
			try {
				_manager.shutdownNoExit(_davConnected);
			}
			catch(FailureException e) {
				e.log();
			}
		}

		File lockFile = new File(_path + LOCKFILE);
		if(!lockFile.delete()) {
			_debug.warning(
					"Lock-Datei konnte nicht gelöscht werden. Bitte löschen Sie " + "die Datei " + lockFile + "."
			);
		}
		disconnectFromDav();
		_davConnected = false;
	}

	/**
	 * Zeigt an ob die SWE vollständig hochgefahren wurde.
	 *
	 * @return	<code>true</code> wenn die SWE bereit zur Annahme von Aufträgen ist. <code>false</code> sonst.
	 */
	public boolean isReady() {
		if(_manager != null) {
			return _manager.isReady();
		}
		else {
			return false;
		}
	}
}
