/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.client;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.kernsoftware.ConnectionManager;
import de.bsvrz.sys.funclib.losb.ringbuffer.RingBuffer;
import jdk.nashorn.internal.ir.ThrowNode;

/**
 * Klasse die den Empfang von Daten über den Datenverteiler übernimmt. Eingehende Daten werden nur gespeichert, wenn diese einen Datenteil haben. Außerdem
 * müssen Anfrage-Id und Sender mit den im Konstruktor übergebenen Objekten übereinstimmen.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public abstract class AnswerReceiver implements ClientReceiverInterface {

	/** Initialisierung Ringpuffer */
	static private final int INITIAL_RINGBUFFER_SIZE = 250;

	/** Verbindung zum Datenverteiler. */
	protected ClientDavInterface dav;

	/** Skriptobjekt, auf das sich die Antwort bezieht. */
	protected SystemObject object;

	/** Eingegangene Daten */
	protected RingBuffer<ResultData> incoming;

	/** Angemeldete Atg und Aspekt. */
	protected DataDescription dataDescription;

	/** Anfrage-Id. */
	protected long requestId;

	/** Objekt das die Anfrage, auf die sich diese Antwort bezieht, gesendet hat. */
	protected SystemObject receiver;

	/** Maximale Wartezeit, in der eine Antwort eingehen muss. */
	private final long _maximalWaitingTime;

	/**
	 * Erzeugt den Empfänger. Empfänger meldet sich sofort als Empfänger auf Daten an.
	 *
	 * @param dav       Verbindung zum Datenverteiler
	 * @param object    Systemobjekt, dessen Daten empfangen werden sollen
	 * @param atg       Pid der Attributgruppe auf die sich der Empfänger anmeldet.
	 * @param asp       Pid des Aspekts auf den sich der Empfänger anmeldet.
	 * @param receiver  Objekt das als Absender im Datensatz eingetragen wist.
	 * @param requestId Anfrage-Id.
	 * @param timeout   Maximale Wartezeit, in der eine Antwort eingehen muss. <code>0</code> ist die unbegrenzte Wartezeit.
	 */
    public AnswerReceiver(
		    ClientDavInterface dav, SystemObject object, String atg, String asp,
		    SystemObject receiver, long requestId, long timeout) {
		this._maximalWaitingTime = timeout;
		this.object = object;
		this.requestId = requestId;
		this.dav = dav;
		this.receiver = receiver;
		incoming = new RingBuffer<ResultData>(INITIAL_RINGBUFFER_SIZE, RingBuffer.UNBOUNDED_SIZE);

		DataModel model = dav.getDataModel();
		this.dataDescription = new DataDescription(model.getAttributeGroup(atg), model.getAspect(asp));
    }

	protected void subscribe(final ReceiverRole role) {
		try {
			ConnectionManager.subscribeReceiver(this.dav, this, this.object, dataDescription, ReceiveOptions.normal(), role);
		}
		catch(FailureException e) {
			// kann nicht geworfen werden, da ein neu erzeugtes Objekt angemeldet wird...
			throw new AssertionError(e);
		}
	}


	/**
	 * Zeigt an, ob Antwortdaten vorliegen.
	 *
	 * @return True, Falls antwortdaten vorliegen.
	 */
	public boolean hasData() {
		return incoming.isEmpty();
	}

	/**
	 * Meldet den Empfang von Daten ab.
	 *
	 * @return Immer <code>true</code>
	 */
    public boolean unsubscribe() {
    	ConnectionManager.unsubscribeReceiver(dav, this, object, dataDescription);
    	return true;
	}

	/** @see ClientReceiverInterface#update(ResultData[]) */
	public void update(ResultData[] results) {
		for(ResultData rd : results) {
			if(rd.getData() != null) {
				update(rd);
			}
		}
	}

	/**
	 * Wird für jeden empfangenen Datensatz von {@link #update(ResultData[])} aufgerufen.
	 *
	 * @param result Eingehender Datensatz. Enthält Daten, d.h. <code>rd.getData()</code> ist nicht <code>null</code>.
	 */
	protected abstract void update(ResultData result);

	/**
	 * Liefert einen Antwortdatensatz. Blockiert, bis Antwortdaten vorliegen oder bis zum übergebenen gesetzten Timeout.
	 *
	 * @param timeout Wartezeit. <code>0</code> bedeutet unbegrenzte Wartezeit.
	 *
	 * @return Antwortdatensatz oder null bei Fehler.
	 *
	 * @throws InterruptedException Warten auf Antwortdatensatz wurde unterbrochen.
	 */
	public ResultData take(long timeout) throws InterruptedException {
		if(timeout > 0) {
			return incoming.pop(timeout);
		}
		else {
			return incoming.pop();
		}
	}

	/**
	 * Liefert einen Antwortdatensatz. Blockiert, bis Antwortdaten vorliegen oder bis zum im Konstruktor gesetzten Timeout.
	 *
	 * @return Antwortdatensatz oder null bei Fehler.
	 *
	 * @throws InterruptedException Warten auf Antwortdatensatz wurde unterbrochen.
	 */
	public ResultData take() throws InterruptedException {
		if(_maximalWaitingTime > 0) {
			return incoming.pop(_maximalWaitingTime);
		}
		else {
			return incoming.pop();
		}
	}
}
