/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.manager;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.losb.datk.AtgAnswer;
import de.bsvrz.sys.funclib.losb.datk.AtlExtra;
import de.bsvrz.sys.funclib.losb.datk.AtlMeta;
import de.bsvrz.sys.funclib.losb.datk.PidScript;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.exceptions.SenderException;
import de.bsvrz.sys.funclib.losb.kernsoftware.SimpleSender;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

/**
 * Klasse zum Senden einer Antwort auf eine Skript-bezogene Anfrage. Die Klasse verwendet {@link de.bsvrz.sys.funclib.losb.kernsoftware.SimpleSender} um ein
 * Skriptobjekt auf den Datenverteiler zu legen. Der Sendevorgang wird als Quelle durchgeführt.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
@SuppressWarnings("deprecation")
public class ScriptAnswer {

	/**
	 * Sendet eine Fehlermeldung an den Auftraggeber. Wartet, bis die Daten gesendet wurden oder es zu einem Sendefehler kommt. Der Sendevorgang wird als Quelle
	 * durchgeführt.
	 *
	 * @param dav       Verbindung zum Datenverteiler
	 * @param script    Skriptobjekt
	 * @param client    Auftraggeber, an den die Fehlermeldung gesendet wird
	 * @param requestId AnfrageId
	 * @param errorMsg  Zu übertragende Fehlermeldung
	 *
	 * @throws FailureException Falls nicht gesendet werden konnte.
	 */
	public static void sendError(ClientDavInterface dav, SystemObject script, SystemObject client, long requestId, String errorMsg) throws FailureException {
		try {
			sendScriptError(dav, script, client, requestId, errorMsg);
		}
		catch(ConfigurationException e) {
			logError(dav, script, client, requestId, ErrorMessage.COMMUNICATION, e, LoggerException.ERROR);
		}
		catch(RuntimeException e) {
			logError(dav, script, client, requestId, e.getMessage(), e, LoggerException.WARNING);
		}
		catch(SenderException e) {
			e.log();
			throw new FailureException(e);
		}
	}

	/**
	 * Sendet eine Fehlermeldung an den Auftraggeber. Wartet, bis die Daten gesendet wurden oder es zu einem Sendefehler kommt. Der Sendevorgang wird als Quelle
	 * durchgeführt.
	 *
	 * @param dav       Verbindung zum Datenverteiler
	 * @param script    Skriptobjekt
	 * @param client    Auftraggeber, an den die Fehlermeldung gesendet wird
	 * @param requestId AnfrageId
	 * @param errorMsg  Zu übertragende Fehlermeldung
	 *
	 * @throws ConfigurationException
	 * @throws SenderException
	 * @throws RuntimeException
	 */
	private static void sendScriptError(ClientDavInterface dav, SystemObject script, SystemObject client, long requestId, String errorMsg)
			throws ConfigurationException, SenderException, RuntimeException {
		DataModel model = dav.getDataModel();

		AttributeGroup atg = model.getAttributeGroup(PidScript.atgAnswer);
		Aspect asp = model.getAspect(PidScript.aspIs);

		DataDescription dd = new DataDescription(atg, asp);

		Data data = AtgAnswer.build(dav, new AtlExtra(client, requestId, ScriptRelatedOperation.NACK_FLAG, ""), new AtlMeta(), errorMsg);

		//senden
		SimpleSender.sourceWait(dav, script, dd, data, 60 * 1000);
	}

	/**
	 * Legt den Quelltext eines Skripts auf den Datenverteiler. Wartet, bis die Daten gesendet wurden oder es zu einem Sendefehler kommt. Der Sendevorgang wird als
	 * Quelle durchgeführt.
	 *
	 * @param dav       Verbindung zum Datenverteiler
	 * @param script    Skript
	 * @param client    Auftraggeber
	 * @param requestId AnfrageId
	 * @param meta      Metainformationen des Skripts
	 *
	 * @param extra
	 * @return Zeigt ob gesendet wurde. Wenn nicht gesendet wurde, dann war die Sendesteuerung negativ.
	 *
	 * @throws FailureException Falls nicht gesendet werden konnte.
	 */
	public static boolean sendSource(ClientDavInterface dav, SystemObject script, SystemObject client, long requestId, AtlMeta meta, final AtlExtra extra)
			throws FailureException {
		try {
			DataModel model = dav.getDataModel();

			//atg & asp der Daten
			AttributeGroup atg = model.getAttributeGroup(PidScript.atgAnswer);
			Aspect asp = model.getAspect(PidScript.aspIs);
			DataDescription dd = new DataDescription(atg, asp);

			//Daten eintragen
			Data data = AtgAnswer.build(dav, extra, meta, "");

			return SimpleSender.sourceWait(dav, script, dd, data, 60 * 1000);
		}
		catch(ConfigurationException e) {
			logError(dav, script, client, requestId, ErrorMessage.COMMUNICATION, e, LoggerException.ERROR);
		}
		catch(RuntimeException e) {
			logError(dav, script, client, requestId, e.getMessage(), e, LoggerException.WARNING);
		}
		catch(SenderException e) {
			e.log();
			throw new FailureException(e);
		}
		return false;
	}

	/**
	 * Legt Metainformationen des Skripts auf den Datenverteiler. Wartet, bis die Daten gesendet wurden oder es zu einem Sendefehler kommt. Der Sendevorgang wird
	 * als Quelle durchgeführt.
	 *
	 * @param dav       Verbindung zum Datenverteiler
	 * @param script    Skriptobjekt.
	 * @param atgAnswer Antwortdaten
	 *
	 * @throws FailureException Falls nicht gesendet werden konnte.
	 */
	public static void sendAnswer(ClientDavInterface dav, SystemObject script, AtgAnswer atgAnswer) throws FailureException {
		try {
			DataModel model = dav.getDataModel();

			//atg & asp der Daten
			AttributeGroup atg = model.getAttributeGroup(PidScript.atgAnswer);
			Aspect asp = model.getAspect(PidScript.aspIs);
			DataDescription dd = new DataDescription(atg, asp);

			//ack-Flag setzten und Daten eintragen
			atgAnswer.atlExtra.opCode = ScriptRelatedOperation.ACK_FLAG;

			SimpleSender.sourceWait(dav, script, dd, atgAnswer.toData(dav), 60 * 1000);
		}
		catch(ConfigurationException e) {
			logError(dav, script, atgAnswer.atlExtra.sender, atgAnswer.atlExtra.requestId, ErrorMessage.COMMUNICATION, e, LoggerException.ERROR);
		}
		catch(RuntimeException e) {
			logError(dav, script, atgAnswer.atlExtra.sender, atgAnswer.atlExtra.requestId, e.getMessage(), e, LoggerException.WARNING);
		}
		catch(SenderException e) {
			e.log();
			throw new FailureException(e);
		}
	}

	/**
	 * Loggt einen Fehler, sendet die Fehlermeldung an den Empfänger und wirft eine FailureException.
	 *
	 * @param dav      Verbindung zum Datenverteiler
	 * @param script   Skript, an das die Fehlermledung gehängt wird.
	 * @param receiver Empfänger (wird in der Attributliste {@link PidScript#atlExtendedInformation} eingetragen.
	 * @param id       Id der Nachricht
	 * @param msg      Nachricht
	 * @param cause    Grund der Ausnahme.
	 * @param logLevel Log-Level. {@link de.bsvrz.sys.funclib.losb.exceptions.LoggerException#ERROR}, {@link de.bsvrz.sys.funclib.losb.exceptions.LoggerException#WARNING}oder
	 *                 {@link de.bsvrz.sys.funclib.losb.exceptions.LoggerException#OTHER}
	 *
	 * @throws FailureException Wird immer geworfen.
	 */
	static boolean logError(ClientDavInterface dav, SystemObject script, SystemObject receiver, long id, String msg, Throwable cause, int logLevel)
			throws FailureException {
		/*
		 * Reihenfolge ist wichtig!
		 * ERST Fehler auf der Debugausgabe ausgeben (dabei kann nichts schiefgehen)
		 * DANN Fehler an Auftraggeber senden - das kann ggf. fehlschlagen - und dann sollte zumindest im Logfile zuerst der ursprüngliche Fehler stehen!
		 * DANACH die Exception werfen
		 */
		msg = id + ": " + msg;
		LoggerException.log(msg, logLevel);
		try {
			sendScriptError(dav, script, receiver, id, msg);
		}
		catch(Exception e) {
			//Fehler wurde geloggt, verschickt konnte er nicht werden...
		}
		//Der auslösende Fehler wird weitergereicht.
		throw new FailureException(msg, cause, logLevel);
	}

	/**
	 * Sendet eine positive Quittung. Der Sendevorgang wird als Quelle durchgeführt. Wartet, bis die Daten gesendet wurden oder es zu einem Sendefehler kommt.
	 *
	 * @param dav       Verbindung zum Datenverteiler
	 * @param script    Skriptobjekt, das auf den Datenverteiler gelegt wird
	 * @param client    Auftraggeber
	 * @param requestId Anfrage-Id
	 *
	 * @throws FailureException Fehler beim Senden.
	 */
	public static void ack(ClientDavInterface dav, SystemObject script, SystemObject client, long requestId) throws FailureException {
		try {
			AttributeGroup atg = dav.getDataModel().getAttributeGroup(PidScript.atgAnswer);
			Aspect asp = dav.getDataModel().getAspect(PidScript.aspIs);

			DataDescription dd = new DataDescription(atg, asp);

			Data answerData = AtgAnswer.build(dav, new AtlExtra(client, requestId, ScriptRelatedOperation.ACK_FLAG, ""), new AtlMeta(), "");

			SimpleSender.sourceWait(dav, script, dd, answerData, 60 * 1000);
		}
		catch(ConfigurationException e) {
			logError(dav, script, client, requestId, ErrorMessage.COMMUNICATION + e.getMessage(), e, LoggerException.ERROR);
		}
		catch(SenderException e) {
			logError(dav, script, client, requestId, e.getMessage(), e, LoggerException.WARNING);
		}
	}
}
