/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.manager.requests;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.exceptions.CriticalParserException;
import de.bsvrz.pua.prot.interpreter.prettyprint.PrettyPrinter;
import de.bsvrz.pua.prot.manager.ManagerErrorMessage;
import de.bsvrz.pua.prot.manager.ScriptAnswer;
import de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface;
import de.bsvrz.pua.prot.util.ProcessingInformation;
import de.bsvrz.pua.prot.util.ProtocolRequestResult;
import de.bsvrz.pua.prot.util.Qualifier;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.AtgAnswer;
import de.bsvrz.sys.funclib.losb.datk.AtgScriptRequest;
import de.bsvrz.sys.funclib.losb.datk.AtlDetail;
import de.bsvrz.sys.funclib.losb.datk.AtlMeta;
import de.bsvrz.sys.funclib.losb.datk.AtlVariable;
import de.bsvrz.sys.funclib.losb.datk.AtlVersion;
import de.bsvrz.sys.funclib.losb.datk.PidScript;
import de.bsvrz.sys.funclib.losb.datk.SerializerUtil;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.*;

/**
 * Auftrag "Skript anlegen oder editieren". Siehe Feinentwurf Kapitel 2.1.2.2.1.1. Erzeugt oder editiert ein Skript.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class CreateScriptRequest extends ScriptRequest {

	/** Debug Ausgaben */
	@SuppressWarnings("hiding")
    private static final Debug _debug = Debug.getLogger();

	/** Verwendeter Algorithmus zum Berechnen der Prüfsumme. */
	public static final String CHECKSUM_ALGORITHM = "MD5";

	/** Auftragsparameter */
	private ResultData _resultData;

	/**
	 * Erzeugt eine Skript-bezogene Anfrage.
	 *
	 * @param dav         Verbindung zum Datenverteiler
	 * @param dataManager Data Manager
	 * @param resultData  Eingangsdaten
	 * @param prettyPrint Soll der Pretty-printer verwendet werden?
	 */
	public CreateScriptRequest(ClientDavInterface dav, DataManagerInterface dataManager, ResultData resultData, boolean prettyPrint) {
		super(dav, dataManager, new ResultData[]{resultData}, prettyPrint);
		this._resultData = resultData;
	}

	/**
	 * Erzeugt oder editiert ein Skriptobjekt & den zugehörigen Quelltext. Dateiname des Quelltextes: Id + Dateiendung .pua Dateiname der Metainformation: Id +
	 * Dateiendung .mta
	 *
	 * @return Objekt mit Informationen über die Auftragsausführung
	 */
	@Override
    public ProtocolRequestResult process() {
		ProtocolRequestResult result = null;
		Data data = _resultData.getData();

		try {
			int versionNumber = 0;
			ArrayList<AtlVariable> atlVariables = new ArrayList<AtlVariable>();

			//Skript parsen
			AtgScriptRequest request = AtgScriptRequest.getJavaObject(data);
			ProcessingInformation pi = dataManager.process(request.atlExtra.source);

			//Überprüfen, ob schon Meta-informationen vorhanden. (Wenn ja: Editieren. Wenn nein: Erzeugen)
			SystemObject scriptObject = _resultData.getObject();
			AtlMeta atlMeta = loadMetaInformation(dataManager.getAbsolutFileNameMeta(scriptObject));

			//Data configMeta = script.getConfigurationData(atg);
			if(atlMeta == null) {
				//1. Erzeugen
				atlMeta = new AtlMeta();
			}
			else {
				//2. EDITIEREN: MetaInfos aus der Konfig holen & Versionshistorie updaten
				final List<AtlVersion> atlVersions = atlMeta.atlVersions;
				versionNumber = atlVersions.size() + 1;
				atlVersions.add(
						new AtlVersion(
								versionNumber, atlMeta.atlBase.date, atlMeta.atlBase.description, atlMeta.atlBase.author
						)
				);
			}

			//Basis-, Detail-, und Variable Attributliste updaten
			atlMeta.atlBase = request.atlBase;
			atlMeta.atlDetail = createDetail(request, dataManager.getAbsolutFileNameMeta(scriptObject), versionNumber);
			for(Qualifier quali : pi.getDataIndices()) {
				atlVariables.add(
						new AtlVariable(
								quali.object.getPid(), quali.attributeGroup.getPid(), quali.getAspect()
						)
				);
			}
			atlMeta.atlVariables = atlVariables;

			//Meta-Informationen persistent ablegen
			_debug.info("Schreibe Daten für: " + scriptObject + ": " + atlMeta);
			saveMetaInfoToDisk(dataManager.getAbsolutFileNameMeta(scriptObject), atlMeta);

			//Quelltext ggf formatieren
			if(printPretty()) {
				try {
					request.atlExtra.source = PrettyPrinter.prettyPrint(request.atlExtra.source);
				}
				catch(Exception e) {
					_debug.warning(ManagerErrorMessage.CAN_NOT_PRETTY_PRINT + "Anfrage : " + requestId, e);
				}
			}
			//Quelletext auf der Platte ablegen
			saveSourceToDisk(dataManager.getAbsolutFileNameSource(scriptObject), request.atlExtra.source);

			//Daten auf Dav legen...
			ScriptAnswer.sendAnswer(dav, scriptObject, new AtgAnswer(request.atlExtra, atlMeta, ""));

			result = new ProtocolRequestResult();
		}
		catch(CriticalParserException e) {
			result = new ProtocolRequestResult(e.getMessage());
		}
		catch(FailureException e) {
			result = new ProtocolRequestResult(e.getMessage());
		}
		catch(NoSuchAlgorithmException e) {
			result = new ProtocolRequestResult(ManagerErrorMessage.INVALID_CHECKSUM_ALGORITHM + CHECKSUM_ALGORITHM + ", " + e.getMessage());
		}
		catch(FileNotFoundException e) {
			result = new ProtocolRequestResult(ErrorMessage.FILE_NOT_FOUND + e.getMessage());
		}
		catch(IOException e) {
			result = new ProtocolRequestResult(ErrorMessage.CAN_NOT_ACCESS_FILE + e.getMessage());
		}

		logError(result);

		return result;
	}


	/**
	 * Speichert Metainformationen in einer Datei.
	 *
	 * @param fileName Dateiname
	 * @param atlMeta  Metainformationen
	 *
	 * @throws FailureException Fehler beim Speichern
	 */
	private static void saveMetaInfoToDisk(String fileName, AtlMeta atlMeta) throws FailureException {
		SerializerUtil.serialize(fileName, atlMeta);
	}

	/**
	 * Speichert den Skriptquelltext
	 *
	 * @param fileName Dateiname
	 * @param source   Quelltext
	 *
	 * @throws IOException Fehler beim Schreiben.
	 */
	private static void saveSourceToDisk(String fileName, String source) throws IOException {
		FileOutputStream fos;
		BufferedWriter out = null;
		try {
			fos = new FileOutputStream(fileName);
			out = new BufferedWriter(
					new OutputStreamWriter(
							fos, Charset.forName("ISO-8859-1")
					)
			);
			out.write(source);
			Util.close(out);
		}
		catch(IOException e) {
			Util.close(out);
			throw e;
		}
	}

	/**
	 * Lädt die zum Skript gehörenden Metainformationen. Der Dateiname der Metainformationen ist die id.
	 *
	 * @param fileName Dateiname der gespeicherten Metainformationen
	 *
	 * @return Metainformationen oder <code>null</code>, falls keine auf dem Datenträger gespeichert war.
	 *
	 * @throws FailureException Fehler beim Zugriff auf die Metainformationen.
	 */
	public static AtlMeta loadMetaInformation(String fileName) throws FailureException {
		AtlMeta meta;

		File f = new File(fileName);
		if(f.exists() && !f.isDirectory()) {
			Object obj = SerializerUtil.deserialize(fileName);
			if(obj instanceof AtlMeta) {
				meta = (AtlMeta)obj;
			}
			else {
				String errorMsg = ErrorMessage.WRONG_OBJECT_SERIALIZED + obj.getClass() + " erwartet: " + AtlMeta.class.getName();
				_debug.warning(errorMsg);
				throw new FailureException(errorMsg, LoggerException.WARNING);
			}
		}
		else {
			meta = null;
		}
		return meta;
	}

	/**
	 * Falls {@link ProtocolRequestResult#isRequestSuccessful()} ist <code>false</code> dann wird der Fehler mittels Debug ausgegeben und an den Auftraggeber
	 * versandt.
	 *
	 * @param result Ergebnis einer Operation.
	 */
	private void logError(ProtocolRequestResult result) {
		if(!result.isRequestSuccessful()) {
			String msg = result.getErrorMsg();
			if(msg == null) msg = "";
			_debug.warning(msg);
			try {
				ScriptAnswer.sendError(dav, script, client, requestId, msg);
			}
			catch(FailureException fe)		//Für die durch sendError ausgelösten Exceptions
			{
				_debug.warning(fe.getMessage());
			}
		}
	}

	/**
	 * Erzeugt die Attributliste {@link PidScript#atlDetailInformation}.
	 *
	 * @param request       Attributgruppe die den Skriptquelltext enthält
	 * @param fileName      Datei, in der die Metainformationen zum Skript abgelegt werden.
	 * @param versionNumber Versionsnummer des Skripts
	 *
	 * @return Attributliste mit ausgefüllten Werten (inkl. Prüfsumme)
	 *
	 * @throws NoSuchAlgorithmException Unbekannter Algorithmus zur Prüfsummenberechnung.
	 * @throws FailureException         Fehler bei der Berechnung der Prüfsumme.
	 */
	private AtlDetail createDetail(AtgScriptRequest request, String fileName, int versionNumber) throws NoSuchAlgorithmException, FailureException {
		byte checksum[] = calcChecksum(request.atlExtra.source);

		AtlDetail atlDetail = new AtlDetail(versionNumber, fileName, checksum);
		return atlDetail;
	}

	/**
	 * Berechnet die Prüfsumme für ein Skript.
	 *
	 * @param source Skriptquelltext.
	 *
	 * @return Prüfsumme.
	 *
	 * @throws FailureException         Fehler bei der Berechnung der Prüfsumme.
	 * @throws NoSuchAlgorithmException Unbekannter Algorithmus zur Prüfsummenberechnung.
	 */
	static public byte[] calcChecksum(String source) throws FailureException, NoSuchAlgorithmException {
		MessageDigest md = MessageDigest.getInstance(CHECKSUM_ALGORITHM);
		md.update(source.getBytes());
		byte[] checksum = md.digest();

		if(checksum == null || checksum.length <= 0) {
			throw new FailureException(ManagerErrorMessage.CAN_NOT_CALCULATE_CHECKSUM, LoggerException.WARNING);
		}
		else {
			return checksum;
		}
	}
}
