/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.manager.taskmanager;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.manager.ManagerErrorMessage;
import de.bsvrz.pua.prot.manager.ProtocolAnswer;
import de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface;
import de.bsvrz.pua.prot.processing.ProcessingErrorMessage;
import de.bsvrz.pua.prot.sender.PuaSender;
import de.bsvrz.pua.prot.util.JobInProgress;
import de.bsvrz.pua.prot.util.JobState;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;

import java.util.*;
import java.util.concurrent.atomic.AtomicLong;

import static de.bsvrz.pua.prot.manager.taskmanager.ThreadElement.Operation;

/**
 * Verwaltet eine feste Anzahl von {@link de.bsvrz.pua.prot.manager.taskmanager.ThreadElement} Objekten. Diese Threads werden über die Protokoll-Id
 * identifiziert. ThreadPool führt eine Liste mi {@link java.lang.ThreadGroup} kann nicht verwendet werden, da die Anzahl der Threads nicht eingestellt werden
 * kann. Außerdem führt ThreadPool eine Liste mit ausstehenden Aufträgen, die auf einen frei werdenen Thread verteilt werden.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ThreadPool {

	/** Wartezeit nach der überprüft wird, ob sich alle Threads beendet haben. */
	private static final long TIMEOUT = 500;

	/** Debug Ausgabe */
	private static final Debug _debug = Debug.getLogger();

	/** Verbindung zum Datenverteiler. */
	private ClientDavInterface _dav;

	/** Konfigurationsverantwortlicher für Anfragen an das Archivsystem. */
	private ConfigurationObject _configAuth;

	/** Warteschlange für Anfragen. */
	private ArrayList<TaskManagerRequest> _queries;

	/** Zeigt an ob die Warteschlange für Anfragen benutzt werden soll. */
	private boolean _useQuery;

	/** Maximale Größe der Warteschlange für Aufträge zur Protokollerstellung. */
	private int _maxQueries;

	/** Objekt zur Thread Verwaltung */
	private ThreadManager _threadManager;

	/**
	 * Zähler zur Erzeugung von eindeutigen Auftrags-Ids (für diese PuA-Instanz). Im Gegensatz zu Protokoll-Ids müssen diese
	 * IDs nur für die aktuelle PuA-Instanz eindeutig sein, es bietet aich also an, einfach einen Int-Wert hichzuzählen.
	 *
	 * Da für Protokoll-Erstellen-Aufträge aus Kompatibilitätsgründen die Protokoll-Id mit der Auftrags-Id identisch sein muss,
	 * werden für Protokoll-IDs nur die 4 niederwertigsten Bits eines Longs verwendet.
	 *
	 * Folglich bleiben, um keine Kollisionen zwischen Protokoll-ID und Auftrags-ID zu erzeugen, für sonstige Aufträge nur IDs
	 * übrig bei denen mindestens ein höherwertiges Bit gesetzt ist. Dies erledigt die Methode generateJobid.
	 */
	private final AtomicLong jobIdCounter = new AtomicLong();

	/**
	 * Erzeugt den ThreadPool. Es werden sofort alle Threads erzeugt. Es wird solange gewartet, bis alle Threads gestartet wurden.
	 *
	 * @param dav
	 * @param maxProtocols Anzahl der Threads die gleichzeitig laufen dürfen
	 * @param dataManager  Datenmanager. Ermöglicht das Interpretieren des Skripts.
	 * @param maxQueries   Maximale Größe der Warteschlange für Anufträge zur Protokollerstellung.
	 * @param maxDelay     Maximale Zeitdauer in ms, für die die Protokollerstellung angehalten werden kann.
	 * @param timeout      Anzahl Minuten ohne Antwort vom Client, nach der die Übertragung getrennt wird
	 * @param configAuth   Konfigurationsverantwortlicher für Anfragen an das Archivsystem.
	 * @throws FailureException Keine Verbindung zum Datenverteiler; Erzeugen aller Threads scheitert.
	 */
	public ThreadPool(ClientDavInterface dav, int maxProtocols, DataManagerInterface dataManager, int maxQueries, long maxDelay, final int timeout, ConfigurationObject configAuth)
			throws FailureException {
		this._dav = dav;

		this._configAuth = configAuth;
		this._maxQueries = maxQueries;
		//Thread-Verwaltung anlegen
		_threadManager = new ThreadManager(maxProtocols);

		if(maxQueries == 0) {
			_useQuery = false;
		}
		else {
			_useQuery = true;
		}
		_queries = new ArrayList<TaskManagerRequest>();

		for(int i = 0; i < maxProtocols; i++) {
			ThreadElement thread = new ThreadElement(this, dataManager, maxDelay, timeout);
			thread.start();
			_threadManager.addInactiveThread(thread);
		}

		if(!_threadManager.hasInactiveThreads()) throw new FailureException("ThreadPool: " + ManagerErrorMessage.NO_THREADS, LoggerException.WARNING);

		/**
		 * Der folgende Code ist nötig um sicherzugehen, dass alle Threads gestartet sind, bevor
		 * auf sie zugegriffen wird.
		 */
		_debug.finest("Starte Threads.");
		_threadManager.waitTillLaunched();
	}


	/**
	 * Aktiviert einen Thread, und übergibt ihm die Auftragsparameter. Erezugt einen Sender zum Versenden der Datensätze.
	 *
	 * @param requestId  Anfrage-Id. Id, die die anfragende Applikation vergeben hat.
	 * @param pp         Auftragsparameter
	 * @param client     Auftraggeber, an den das Ergebnis gesendet wird.
	 * @param protocolId Protokoll-Id
	 * @param operation  Auftragsart. {@link  de.bsvrz.pua.prot.manager.taskmanager.ThreadElement.Operation#CREATE_PROTOCOL}, {@link  de.bsvrz.pua.prot.manager.taskmanager.ThreadElement.Operation#SEND_SAVED_PROTOCOL} oder {@link
	 *                   de.bsvrz.pua.prot.manager.taskmanager.ThreadElement.Operation#SEND_UNREAD_PROTOCOL}
	 *
	 * @param jobId      Auftrags-Id
	 * @param startTime
	 * @param initialFlowControl
	 * @return <code>false</code>falls alle Threads bereits beschäftigt sind oder kein Sender erzeugt werden konnte.<code>true</code> sonst.
	 */
	private synchronized boolean activateThread(long requestId, ProcessingParameter pp, SystemObject client, long protocolId, Operation operation, final long jobId, final long startTime, final long initialFlowControl) {
		ThreadElement thread = _threadManager.getAvailableThread(jobId);
		if(thread != null) {
			try {
				_debug.finest("Aktiviere Thread " + protocolId);
				switch(operation) {
					case CREATE_PROTOCOL:
						thread.activateCreateProtocol(_dav, client, requestId, pp, protocolId, jobId, startTime, initialFlowControl);
						break;
					case SEND_SAVED_PROTOCOL:
						thread.activateGetProtocol(_dav, client, requestId, protocolId, jobId, startTime, initialFlowControl);
						break;
					case SEND_UNREAD_PROTOCOL:
						thread.activateGetUnreadProtocol(_dav, client, requestId, protocolId, jobId, startTime, initialFlowControl);
						break;
					default:
						_debug.warning("Ungültige Operation: " + operation);
				}
				return true;
			}
			catch(FailureException e) {
				e.log();
				return false;
			}
		}
		else {
			_debug.finest("Kein Thread für die Anfrage " + protocolId + " frei!");
			try {
				ProtocolAnswer.sendError(ProtocolAnswer.sender(_dav, client), requestId, ManagerErrorMessage.TOO_MANY_PROTOCOLS);
			}
			catch(FailureException e) {
				_debug.info("Anfrage von " + client + ": " + requestId + ", " + ManagerErrorMessage.TOO_MANY_PROTOCOLS);
			}
			return false;
		}
	}

	/**
	 * Entfernt einen Thread aus der Liste der aktiven Threads. Meldet den Sender ab. Falls Aufträge in der Warteschlange liegen, wird der erste aus der
	 * Warteschlange an den Thread <code> processor</code> vergeben. Diese Methode muss von dem {@link ThreadElement} aufgerufen werden, nachdem die
	 * Protokollerstellung abgeschlossen wurde.
	 *
	 * @param processor Thread der entfernt werden soll.
	 */
	synchronized void finished(ThreadElement processor) {
		if(processor != null) {
			if(!_queries.isEmpty()) {
				TaskManagerRequest tmr;
				tmr = _queries.remove(0);
				try {
					_threadManager.changeId(processor, tmr.jobId);
					switch(tmr.operation) {
						case CREATE_PROTOCOL:
							processor.activateCreateProtocol(_dav, tmr.client, tmr.requestId, tmr.pp, tmr.protocolId, tmr.jobId, tmr.startTime, tmr.initialFlowControl);
							break;
						case SEND_SAVED_PROTOCOL:
							processor.activateGetProtocol(_dav, tmr.client, tmr.requestId, tmr.protocolId, tmr.jobId, tmr.startTime, tmr.initialFlowControl);
							break;
						case SEND_UNREAD_PROTOCOL:
							processor.activateGetUnreadProtocol(_dav, tmr.client, tmr.requestId, tmr.protocolId, tmr.jobId, tmr.startTime, tmr.initialFlowControl);
							break;
						default:
							_debug.warning("Ungültige Operation: " + tmr.operation);
					}
				}
				catch(FailureException e) {
					e.log();
					_threadManager.deactivate(processor);
				}
			}
			else {
				_threadManager.deactivate(processor);
			}
		}
	}


	/**
	 * Zeigt an ob weitere Anfragen angenommen werden können. Dazu wird überprüft, ob die Warteschlange für eingehende Anfragen gefüllt ist und die Threads alle
	 * beschäftigt sind.
	 *
	 * @return	<code>true</code> falls die Warteschlange bereits voll ist. <code>false</code> sonst.
	 */
	public synchronized boolean busy() {
		if(!_useQuery) {
			return !_threadManager.hasInactiveThreads();
		}
		else {
			return _queries.size() == _maxQueries;
		}
	}


	/** Bricht alle Aufträge ab und beendet alle Threads. Anschließend kann mit {@link #isTerminated()} überprüft werden, ob die Threads beendet wurden. */
	public void killAll() {
		synchronized (this) {
			_queries.clear();
		}
		_threadManager.killAll(TIMEOUT);
	}

	/** @return Anzahl der aktiven und inaktiven Threads. */
	public int alive() {
		return _threadManager.alive();
	}

	/** @return Returns the cofigAuth. */
	public ConfigurationObject getConfigAuth() {
		return _configAuth;
	}


	/**
	 * Legt einen Auftrag zur Protokollerstellung in der Warteschlange ab.
	 *
	 * @param requestId  Anfrage-Id. Id, die die anfragende Applikation vergeben hat.
	 * @param pp         Auftragsparameter
	 * @param client     Auftraggeber, an den das Ergebnis gesendet wird.
	 * @param protocolId Protokoll-Id
	 *
	 * @param initialFlowControl
	 * @return <code>true</code> falls der Auftrag angenommen wurde.
	 *
	 * @throws FailureException Fehler beim Senden.
	 */
	public boolean storeCreateQuery(long requestId, ProcessingParameter pp, SystemObject client, long protocolId, final long initialFlowControl) throws FailureException {
		return storeQuery(requestId, pp, client, protocolId, Operation.CREATE_PROTOCOL, initialFlowControl);
	}

	/**
	 * Legt einen Auftrag zum ausgeben eines gespeicherten Protokolls in der Warteschlange ab.
	 *
	 * @param requestId  Anfrage-Id. Id, die die anfragende Applikation vergeben hat.
	 * @param client     Auftraggeber, an den das Ergebnis gesendet wird.
	 * @param protocolId Protokoll-Id des Protokolls, das abgerufen werden soll.
	 *
	 * @param initialFlowControl
	 * @return <code>true</code> falls der Auftrag angenommen wurde.
	 *
	 * @throws FailureException Fehler beim Senden.
	 */
	public boolean storeGetProtocolQuery(long requestId, SystemObject client, long protocolId, final long initialFlowControl) throws FailureException {
		return storeQuery(requestId, null, client, protocolId, Operation.SEND_SAVED_PROTOCOL, initialFlowControl);
	}

	/**
	 * Legt einen Auftrag zum ausgeben eines noch nicht abgelegten Protokolls in der Warteschlange ab.
	 *
	 * @param requestId  Anfrage-Id. Id, die die anfragende Applikation vergeben hat.
	 * @param client     Auftraggeber, an den das Ergebnis gesendet wird.
	 * @param protocolId Protokoll-Id des Protokolls, das abgerufen werden soll.
	 *
	 * @param initialFlowControl
	 * @return <code>true</code> falls der Auftrag angenommen wurde.
	 *
	 * @throws FailureException Fehler beim Senden.
	 */
	public boolean storeGetUnreadProtocolQuery(long requestId, SystemObject client, long protocolId, final long initialFlowControl) throws FailureException {
		return storeQuery(requestId, null, client, protocolId, Operation.SEND_UNREAD_PROTOCOL, initialFlowControl);
	}

	/**
	 * Legt den Auftrag in der Warteschlange ab.
	 *
	 * @param requestId  Anfrage-Id. Id, die die anfragende Applikation vergeben hat.
	 * @param pp         Auftragsparameter
	 * @param client     Auftraggeber, an den das Ergebnis gesendet wird.
	 * @param protocolId Protokoll-Id
	 * @param operation  Auftragsart. {@link de.bsvrz.pua.prot.manager.taskmanager.ThreadElement.Operation#CREATE_PROTOCOL}, {@link de.bsvrz.pua.prot.manager.taskmanager.ThreadElement.Operation#SEND_SAVED_PROTOCOL} oder {@link
	 *                   de.bsvrz.pua.prot.manager.taskmanager.ThreadElement.Operation#SEND_UNREAD_PROTOCOL}
	 *
	 * @param initialFlowControl
	 * @return <code>true</code> falls der Auftrag angenommen wurde.
	 *
	 * @throws FailureException Fehler beim Senden.
	 */
	private synchronized boolean storeQuery(long requestId, ProcessingParameter pp, SystemObject client, long protocolId, Operation operation, final long initialFlowControl)
			throws FailureException {
		final long jobId = operation == Operation.CREATE_PROTOCOL ? protocolId : generateJobId();
		long startTime = System.currentTimeMillis();
		PuaSender sender = ProtocolAnswer.sender(_dav, client);
		if(_threadManager.hasInactiveThreads()) {
			//es gibt idle Thread											-> Direkt an Thread weiterleiten

			// Job-ID direkt an Empfänger senden. Ist notwendig, da später die gesammte Kommunikation über diese abgewickelt wird.
			ProtocolAnswer.sendJobId(sender, requestId, jobId);

			activateThread(requestId, pp, client, protocolId, operation, jobId, startTime, initialFlowControl);

			return true;
		}
		else if(!_useQuery || busy()) {
			//alle Threads belegt, Warteschlange soll nicht verwendet werden oder: Warteschlange voll-> Fehler
			_debug.fine(ManagerErrorMessage.TOO_MANY_PROTOCOLS);
			ProtocolAnswer.sendError(sender, requestId, ManagerErrorMessage.TOO_MANY_PROTOCOLS);
			return false;
		}
		else {
			//alle Threads belegt, Warteschlange verwenden.					-> Auftrag in Queue ablegen
			//Warteschlange ist nicht vollständig belegt

			// Job-ID direkt an Empfänger senden. Ist notwendig, da später die gesammte Kommunikation über diese abgewickelt wird.

			// Die Job-ID wird möglichst schnell (ggf. an der Warteschlange vorbei) an den Empfänger gesendet, damit dieser die
			// Job-ID z.B. zur Flusskontrolle oder zum Abbrechen des Auftrags möglichst schnell verwenden kann
			ProtocolAnswer.sendJobId(sender, requestId, jobId);

			TaskManagerRequest tmr = new TaskManagerRequest(requestId, pp, client, protocolId, operation, jobId, startTime, initialFlowControl);
			_queries.add(tmr);

			return true;
		}

	}

	/**
	 * Bricht einen Auftrag in Warteschlange ab
	 * @param jobId Auftrags-Id des gesuchten Threads
	 * @return der abgebrochene Task falls vorhanden
	 */
	public synchronized QueuedThread abortQueuedTask(final long jobId) {
		for(Iterator<TaskManagerRequest> iterator = _queries.iterator(); iterator.hasNext(); ) {
			final TaskManagerRequest query = iterator.next();
			if(query.jobId == jobId){
				iterator.remove();
				return new QueuedThread(query.requestId, query.protocolId, query.jobId, query.operation, query.client);
			}
		}
		return null;
	}

	/**
	 * Generiert eine (für diesen Prozess) eindeutige Auftrags-Id.
	 *
	 * Da für Protokoll-Erstellen-Aufträge aus Kompatibilitätsgründen die Protokoll-Id mit der Auftrags-Id identisch sein muss,
	 * werden für Protokoll-IDs nur die 4 niederwertigsten Bits eines Longs verwendet.
	 *
	 * Diese Methode wird nur für andere Auftrags-IDs verwendet.
	 *
	 * Folglich bleiben, um keine Kollisionen zwischen Protokoll-ID und Auftrags-ID zu erzeugen, für sonstige Aufträge nur IDs
	 * übrig bei denen mindestens ein höherwertiges Bit gesetzt ist
	 *
	 * @return Auftrags-Id
	 */
	private long generateJobId() {
		long tmp = jobIdCounter.incrementAndGet();
		tmp &= 0x7FFFFFFFFFFFFFFFl; // Vorzeichen-Bit löschen
		tmp |= 0x4000000000000000l; // Ein Bit setzen wegen Kollisionsfreiheit zu Protokoll-IDs
		// Überlauf (und daher nicht-Eindeutigkeit) alle 0x4000000000000000 ~ 4*10^18 Aufträge. -> Vernachlässigbar
		return tmp;
	}

	/**
	 * Gibt eine Liste mit allen aktuell wartenden und in Bearbeitung befindlichen Aufträgen zurück
	 * @return Liste mit Autrags-Objekten
	 * @throws FailureException
	 */
	public synchronized ArrayList<JobInProgress> getJobList() throws FailureException {
		final ArrayList<JobInProgress> jobsInProgress = new ArrayList<JobInProgress>();
		for(TaskManagerRequest query : _queries) {
			jobsInProgress.add(new JobInProgress(query.jobId, query.operation.getCode(), JobState.InQueue, (byte) 0, query.startTime, query.client.getId(), query.pp != null ? query.pp : query.protocolId));
		}
		for(Long jobId : getActiveProtocols()) {
			ThreadElement thread = getThread(jobId);
			if(thread != null){
				JobInProgress jobInProgress = thread.toJobInProgress();
				if(jobInProgress != null){
					jobsInProgress.add(jobInProgress);
				}
			}
		}
		return jobsInProgress;
	}

	/**
	 * Kapselt Daten für die Aufräge Protokollerstellung und gespeichertes Protokoll abrufen..
	 *
	 * @author Martin Hilgers
	 */
	private class TaskManagerRequest {

		/** Anfrage Id (vom Client) */
		public final long requestId;

		/** Auftragsparameter */
		public final ProcessingParameter pp;

		/** Auftraggeber */
		public final SystemObject client;

		/** Protokoll-id */
		public final long protocolId;

		/**
		 * Eindeutige Auftrags-ID
		 */
		public final long jobId;

		/**
		 * Operation: Protokoll erstellen; erstelltes Protokoll senden; noch nicht abgerufenes Protokoll senden. Kann die Werte {@link Operation#CREATE_PROTOCOL},
		 * {@link Operation#SEND_SAVED_PROTOCOL} und {@link Operation#SEND_UNREAD_PROTOCOL} annehmen.
		 */
		public final Operation operation;

		/**
		 * Startzeitpunkt
		 */
		public long startTime;

		/**
		 * Startdatensatznummer der Flusskontrolle
		 */
		public long initialFlowControl;

		/**
		 * @param requestId  Anfrage-Id
		 * @param pp         Auftragsparameter. Kann beim Auftrag 'senden eines gespeicherten Protokolls' <code>null</code> sein.
		 * @param client     Auftraggeber
		 * @param protocolId Protokoll-Id
		 * @param operation  Auftragsart. {@link de.bsvrz.pua.prot.manager.taskmanager.ThreadElement.Operation#CREATE_PROTOCOL}, {@link de.bsvrz.pua.prot.manager.taskmanager.ThreadElement.Operation#SEND_SAVED_PROTOCOL} oder {@link
*                   de.bsvrz.pua.prot.manager.taskmanager.ThreadElement.Operation#SEND_UNREAD_PROTOCOL}
		 * @param jobId      Auftrags-Id
		 * @param startTime  Startzeitpunkt
		 * @param initialFlowControl
		 */
		private TaskManagerRequest(long requestId, ProcessingParameter pp, SystemObject client, long protocolId, Operation operation, final long jobId, final long startTime, final long initialFlowControl) {
			this.requestId = requestId;
			this.pp = pp;
			this.client = client;
			this.protocolId = protocolId;
			this.operation = operation;
			this.jobId = jobId;
			this.startTime = startTime;
			this.initialFlowControl = initialFlowControl;
		}

		/** @see java.lang.Object#toString() */
		@Override
        public String toString() {
			return requestId + ": Protokoll-Id: " + protocolId + ", Operation: " + operation;
		}
	}


	/** Speichert einen Auftrag in der Warteschlange für externe Zwecke */
	public static class QueuedThread {

		private final Long _requestId;

		private final Long _protocolId;

		private long _jobId;

		private final Operation _operation;

		private final SystemObject _client;

		/**
		 * Erstellt eine neue QueuedThread-instanz
		 * @param requestId  Anfrage-ID
		 * @param protocolId Protokoll-ID
		 * @param jobId
		 * @param operation Operation (siehe {@link de.bsvrz.pua.prot.manager.taskmanager.ThreadElement#_operation})
		 * @param client
		 */
		private QueuedThread(final Long requestId, final Long protocolId, final long jobId, final Operation operation, final SystemObject client) {
			_requestId = requestId;
			_protocolId = protocolId;
			_jobId = jobId;
			_operation = operation;
			_client = client;
		}

		public Long getRequestId() {
			return _requestId;
		}

		public Long getProtocolId() {
			return _protocolId;
		}

		public Operation getOperation() {
			return _operation;
		}

		/**
		 * Gibt die eindeutige ID dieses Auftrags zurück. Diese Id wird statt der Protokoll-ID verwendet, da je Protokoll u.U. mehrere
		 * Aufträge gleichzeitig ausgeführt werden können.
		 * @return ID des Auftrags.
		 */
		public long getJobId() {
			return _jobId;
		}

		public SystemObject getClient() {
			return _client;
		}

		@Override
		public String toString() {
			return "QueuedThread{" + "_requestId=" + _requestId + ", _protocolId=" + _protocolId + ", _operation=" + _operation + '}';
		}
	}

	/**
	 * Liste mit den Protokollen die gerade gesendet / erstellt werden.
	 *
	 * @return Liste mit Protokollen. Falls keine Protokolle erstellt werden, wird ein Feld der Länge 0 zurückgegeben.
	 */
	public Long[] getActiveProtocols() {
		return _threadManager.getActiveProtocols();
	}

	/**
	 * Liste mit den Objekten, die momentan in der Warteschlange liegen.Falls keine werte vorliegen, wird ein Feld der Länge 0
	 * zurückgegeben.
	 *
	 * @return Liste QueuedThread-Objekten
	 */
	public synchronized List<QueuedThread> getWaitingProtocols() {
		if(_useQuery) {
			final ArrayList<QueuedThread> list = new ArrayList<QueuedThread>();
			for (TaskManagerRequest tmr : _queries) {
				list.add(new QueuedThread(tmr.requestId, tmr.protocolId, tmr.jobId, tmr.operation, tmr.client));
			}
			return list;
		}
		else {
			return Collections.emptyList();
		}
	}


	/**
	 * Liefert den aktiven Thread mit der gegebenen Auftrags-Id zurück.
	 *
	 * @param jobId Auftrags-Id des gesuchten Threads
	 *
	 * @return Thread mit der gegebenen Auftrags-Id oder null, falls es keinen Thread mit der gegebenen Auftrags-Id gibt.
	 */
	public ThreadElement getThread(long jobId) {
		return _threadManager.getThread(jobId);
	}

	/**
	 * Gibt alle aktiven Threads zurück
	 * @return alle aktiven Threads
	 */
	public ArrayList<ThreadElement> getThreads() {
		return _threadManager.getThreads();
	}

	/**
	 * Zeigt an ob alle Threads beendet wurden.
	 *
	 * @return <code>true</code> falls alle Threads beendet wurden.
	 */
	public boolean isTerminated() {
		return _threadManager.isTerminated();
	}
}

/**
 * Verwaltet Threads. Synchronisiert den Zugriff auf die aktiven und unbeschäftigten Threads.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 */
class ThreadManager {

	/** Debug Ausgabe */
	private static final Debug debug = Debug.getLogger();

	/** Abbildung von Protokoll-Id auf den jeweiligen Thread. Jeder Zugriff auf diese Hashtable muss synchronisiert erfolgen! */
	private Hashtable<Long, ThreadElement> activeThreads;

	/** Liste mit den idle Threads. Jeder Zugriff auf diese Liste muss synchronisiert erfolgen! */
	private ArrayList<ThreadElement> inactiveThreads;

	/** Wird zum Herunterfahren verwendet. Enthält alle Threads, die noch nicht beendet wurden. */
	private ArrayList<ThreadElement> stillAlive;

	/** Zeigt an, dass sich die Threads beenden sollen. */
	private boolean die = false;

	/** @param maxProtocols Anzahl Threads. */
	public ThreadManager(int maxProtocols) {
		activeThreads = new Hashtable<Long, ThreadElement>(maxProtocols);
		inactiveThreads = new ArrayList<ThreadElement>(maxProtocols);
	}

	/**
	 * Ändert die Protokoll-Id, die einem Thread zugeordnet wird.
	 *
	 * @param thread     Thread
	 * @param jobId Neue Job-Id
	 */
	public void changeId(ThreadElement thread, long jobId) {
		if(!die && thread != null) {
			synchronized(activeThreads) {
				activeThreads.remove(thread.getJobId());
				activeThreads.put(jobId, thread);
			}
		}
	}

	/** @return Anzahl der aktiven und unbeschäftigten Threads. */
	public int alive() {
		return activeThreads.size() + inactiveThreads.size();
	}

	/**
	 * Liste mit den Protokollen die gerade gesendet / erstellt werden.
	 *
	 * @return Liste mit Protokollen. Falls keine Protokolle erstellt werden, wird ein Feld der Länge 0 zurückgegeben.
	 */
	public Long[] getActiveProtocols() {
		Long protocols[];
		synchronized(activeThreads) {
			protocols = new Long[activeThreads.size()];
			protocols = activeThreads.keySet().toArray(protocols);
		}
		return protocols;
	}

	/**
	 * Bricht alle Aufträge ab und beendet alle Threads.
	 *
	 * @param timeout Zeitspanne, nach der geprüft wird
	 */
	public void killAll(long timeout) {
		die = true;

		stillAlive = new ArrayList<ThreadElement>(activeThreads.size() + inactiveThreads.size());

		synchronized(activeThreads) {
			for(ThreadElement thread : activeThreads.values()) {
				thread.kill();
				thread.interrupt();
				stillAlive.add(thread);
			}
		}

		synchronized(inactiveThreads) {
			for(ThreadElement thread : inactiveThreads) {
				thread.kill();
				thread.interrupt();
				stillAlive.add(thread);
			}
		}

		try {
			Thread.sleep(timeout);
		}
		catch(Exception e) {
		}

		//prüfen, ob sich alle beendet haben
		StringBuilder sb = null;
		for(int i = 0; i < stillAlive.size(); i++) {
			ThreadElement thread = stillAlive.get(i);
			if(thread.isAlive()) {
				debug.warning(ProcessingErrorMessage.THREAD_STILL_RUNNING + thread.getName());
				if(sb == null) sb = new StringBuilder();
				sb.append((sb.toString().equals("") ? "" : "\n") + ProcessingErrorMessage.THREAD_STILL_RUNNING + thread.getName());
			}
		}
		//anschließend der Vollständigkeit halber (damit alive() korrekte Werte meldet) die Listen löschen:
		//das kann nicht weiter oben gemacht werden, sondern erst nach der Prüfung, ob alles beendet.
		//Grund: Wenn sich die Threads beenden, tragen sie sich wieder in activeThreads ein.
		synchronized(activeThreads) {
			activeThreads.clear();
		}
		synchronized(inactiveThreads) {
			inactiveThreads.clear();
		}

		if(sb != null) debug.fine(sb.toString());
	}

	/** Wartet, bis alle Threads gestartet wurden. Übernimmt nicht das Starten der Threads! */
	public void waitTillLaunched() {
		for(ThreadElement t : inactiveThreads) {
			while(!t.running) {
				try {
					Thread.sleep(500);
				}
				catch(InterruptedException e) {
				}
				debug.fine("Thread " + t + " wird noch initialisiert...");
			}
		}
	}

	/**
	 * Zeigt ob alle Threads Daten aufbereiten.
	 *
	 * @return	<code>true</code> falls es unbeschäftige Threads gibt, die momentan keine Daten aufbereiten.
	 */
	public boolean hasInactiveThreads() {
		return !inactiveThreads.isEmpty();
	}

	/**
	 * Fügt einen unbeschäftigten Thread hinzu.
	 *
	 * @param thread Unbeschäftigter Thread.
	 */
	public void addInactiveThread(ThreadElement thread) {
		if(!die) {
			synchronized(inactiveThreads) {
				inactiveThreads.add(thread);
			}
		}
	}

	/**
	 * Liefert einen unbeschäftigten Thread zurück und entfernt ihn aus {@link #inactiveThreads}.
	 *
	 * @param jobId Auftrags-Id, die dem Thread zugewiesen werden soll.
	 *
	 * @return Thread oder <code>null</code> falls alle Threads beschäftigt sind.
	 */
	public ThreadElement getAvailableThread(long jobId) {
		ThreadElement thread = null;
		synchronized(inactiveThreads) {
			if(!inactiveThreads.isEmpty()) thread = inactiveThreads.remove(0);
		}
		if(thread != null) {
			synchronized(activeThreads) {
				activeThreads.put(jobId, thread);
			}
		}
		return thread;
	}

	/**
	 * Liefert den aktiven Thread mit der gegebenen Auftrags-Id zurück.
	 *
	 * @param jobId Auftrags-Id des gesuchten Threads
	 *
	 * @return Thread mit der gegebenen Auftrags-Id oder null, falls es keinen Thread mit der gegebenen Auftrags-Id gibt.
	 */
	public ThreadElement getThread(long jobId) {
		synchronized(activeThreads) {
			return activeThreads.get(jobId);
		}
	}

	/**
	 * Gibt alle aktiven Threads zurück
	 * @return alle aktiven Threads
	 */
	public ArrayList<ThreadElement> getThreads() {
		synchronized(activeThreads) {
			return new ArrayList<ThreadElement>(activeThreads.values());
		}
	}

	/**
	 * Deaktiviert einen aktiven Thread und markiert ihn als unbeschäftigt.
	 *
	 * @param thread Thread, der deaktiviert wird.
	 */
	public void deactivate(ThreadElement thread) {
		if(thread != null && !die) {
			synchronized(activeThreads) {
				activeThreads.remove(thread.getJobId());
			}
			synchronized(inactiveThreads) {
				inactiveThreads.add(thread);
			}
		}
	}

	/**
	 * Zeigt an ob alle Threads beendet wurden.
	 *
	 * @return <code>true</code> falls alle Threads beendet wurden.
	 */
	public boolean isTerminated() {
		ArrayList<ThreadElement> temp = new ArrayList<ThreadElement>(stillAlive.size());
		for(ThreadElement thread : stillAlive) {
			if(thread.isAlive()) {
				thread.kill();
				thread.interrupt();
				temp.add(thread);
			}
		}
		stillAlive = temp;
		return stillAlive.isEmpty();
	}

	/**
	 * Liefert einen String mit den Anfrage und Protokoll-Ids alle Protokolle, die momentan erstellt werden. Für Debug-Ausgabe.
	 *
	 * @return String mit Infos.
	 */
	protected String getInfo() {
		StringBuilder sb = new StringBuilder();
		for(ThreadElement te : activeThreads.values()) {
			sb.append("Anfrage: " + te.getRequestId() + " -> Protokoll: " + te.getProtocolId() + "\n");
		}
		return sb.toString();
	}
}
