/*
 * Copyright 2014 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.processing;

import de.bsvrz.sys.funclib.debug.Debug;

import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicLong;

/**
 * Basisklasse für Protokolldaten-sendende Klassen, die eine Flusskontrolle unterstützen.
 *
 * Bei jedem erfolgreich versendeten Datensatz muss {@link #datasetSend()} aufgerufen werden um den Zähler der gesendeten Daten hochzuzählen.
 *
 * {@link #flowControl()} kann immer aufgerufen werden bevor ein Datensätze versendet werden würde um den Versand ggf. auszubremsen.
 *
 * @author Kappich Systemberatung
 * @version $Revision$
 */
public abstract class BasicFlowControl implements ProcessingInterface {
	private static final Debug _debug = Debug.getLogger();

	/** Flußkontrolle. Objekt wird zum synchronisieren verwendet. */
	protected final AtomicBoolean _pause = new AtomicBoolean(false);

	/** Maximale Zeitdauer in ms für die die Protokollerstellung angehalten werden kann. */
	protected final long _maxDelay;

	/**
	 * Anzahl der bisher versendeten Datensätze
	 */
	private final AtomicLong _actualDatasets = new AtomicLong(0);

	/**
	 * Anzahl der in voller Geschwindigkeit zu sendenden Datensätze.
	 * Standardmäßig "unendlich" für Clients die keine Flusskontrolle unterstützen.
	 */
	private final AtomicLong _desiredDatasets;

	/**
	 * Anzahl Minuten in Folge, die der Client keine Flusskontrollnachricht geschickt hat
	 */
	private int _timeoutCounter;

	/**
	 * Anzahl Minuten ohne Antwort vom Client, nach der die Übertragung getrennt wird
	 */
	private int _timeout;

	/**
	 * Erzeut eine neue Instanz
	 * @param maxDelay Maximale wartezeit bei manueller Flusskontrolle in ms
	 * @param timeout Anzahl Minuten ohne Antwort vom Client, nach der die Übertragung getrennt wird
	 * @param initialFlowControl Anzahl initial zu sendende Datensätze bis zu automatische Flusskontrolle den Versand bremst
	 */
	public BasicFlowControl(final long maxDelay, final int timeout, final long initialFlowControl) {
		this._maxDelay = maxDelay;
		_timeout = timeout;
		_desiredDatasets = new AtomicLong(initialFlowControl <= 0 ? Long.MAX_VALUE : initialFlowControl);
	}

	/** @see de.bsvrz.pua.prot.processing.ProcessingInterface#pauseProcessing() */
	@Override
	public void pauseProcessing() {
		_pause.set(true);
	}

	/** @see de.bsvrz.pua.prot.processing.ProcessingInterface#resumeProcessing() */
	@Override
	public void resumeProcessing() {
		synchronized(_pause) {
			_pause.set(false);
			_pause.notifyAll();
		}
	}

	/**
	 * Flußkontrolle. Ggf. wird wird gewartet bis das Senden fortgesetzt werden kann.
	 *
	 * @throws InterruptedException Warten wurde unterbrochen.
	 */
	protected void flowControl() throws InterruptedException {
		while(_pause.get()) {
			synchronized(_pause) {
				_debug.info("Aktiviere manuelle Flusskontrolle für", getJobId());
				_pause.wait(_maxDelay);
				_pause.set(false);
				_debug.info("Setze Protokollerstellung fort", getJobId());
			}
		}
		// Automatische Flusskontrolle wenn die gewünschte Datensatzanzahl erreicht ist.
		// Maximal eine Minute warten oder bis eine Flusskontrollnachricht ankommt.
		synchronized(_desiredDatasets) {
			if(_actualDatasets.get() >= _desiredDatasets.get()) {
				_debug.fine("Aktiviere automatische Flusskontrolle für", getJobId());
				_debug.fine("Aktuelle Datensätze", _actualDatasets);
				_debug.fine("Gewünschte Datensätze", _desiredDatasets);
				_timeoutCounter++;
				_desiredDatasets.wait(60000);
				_debug.fine("Setze Protokollerstellung fort", getJobId());
			}
			else {
				_timeoutCounter = 0;
			}
			if(getTimeout() != -1 && _timeoutCounter >= getTimeout()) {
				String warningMessage = "Client " + getClient() + " hat " + _timeoutCounter + " Minuten lang keine Protokolldaten mehr angefordert.";
				warningMessage = warningMessage + " Die Übertragung wurde abgebrochen.";
				_debug.warning(warningMessage);
				abort();
				return;
			}
			if(_timeoutCounter >= 5) {
				String warningMessage = "Client " + getClient() + " hat " + _timeoutCounter + " Minuten lang keine Protokolldaten mehr angefordert.";
				if(getTimeout() != -1) {
					warningMessage = warningMessage + " Nach " + getTimeout() + " Minuten wird die Übertragung abgebrochen.";
				}
				_debug.warning(warningMessage);
			}
		}
	}

	protected abstract Object getClient();

	@Override
	public abstract long getJobId();

	@Override
	public void flowControl(final long numData) {
		_debug.fine("Flusskontrolle für " + getJobId() + " " + numData + " (gesendet: " + _actualDatasets + ")");
		synchronized(_desiredDatasets) {
			_timeoutCounter = 0;
			_desiredDatasets.set(numData);
			_desiredDatasets.notifyAll();
		}
	}

	public void datasetSend(){
		_actualDatasets.incrementAndGet();
	}

	public int getTimeout() {
		return _timeout;
	}
}
