/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.util;

import antlr.collections.AST;
import de.bsvrz.dav.daf.main.config.Attribute;
import de.bsvrz.dav.daf.main.config.AttributeType;
import de.bsvrz.dav.daf.main.config.DoubleAttributeType;
import de.bsvrz.dav.daf.main.config.IntegerAttributeType;
import de.bsvrz.dav.daf.main.config.ReferenceAttributeType;
import de.bsvrz.dav.daf.main.config.StringAttributeType;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.TimeAttributeType;
import de.bsvrz.pua.prot.client.dataobject.PuADataState;
import de.bsvrz.pua.prot.exceptions.CriticalParserException;
import de.bsvrz.pua.prot.exceptions.SemanticErrorException;
import de.bsvrz.pua.prot.functions.ExpressionInterface;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.processing.util.BaseDataSet;
import de.bsvrz.pua.prot.processing.util.ExpressionResultAndState;
import de.bsvrz.pua.prot.processing.util.TempAttribut;
import de.bsvrz.pua.prot.util.attributes.AttributeDescription;
import de.bsvrz.pua.prot.util.attributes.DurationAttributeDescription;
import de.bsvrz.pua.prot.util.attributes.RealAttributeDescription;
import de.bsvrz.pua.prot.util.attributes.RealElement;
import de.bsvrz.pua.prot.util.attributes.TempAttributeDescription;

import java.util.*;

/**
 * Repräsentiert ein Blatt in einem {@link de.bsvrz.pua.prot.util.ExpressionTree} Ein Blatt enthält entweder eine Zeichenkette, Bool, Attribute, Zahlen oder
 * eine Funktion mit Ausdruck
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ExpressionLeaf extends ExpressionTree {

	/** Typ, der in diesem Blatt gespeichert wird. */
	public enum LeafType {

		/** Wert ist ein String */
		STRING,
		/** Wert ist ein Double */
		DOUBLE,
		/** Wert ist ein Long */
		LONG,
		/** Wert ist ein Attribut */
		ATTRIBUTE,
		/** Wert ist ein Bool */
		BOOL,
		/** Wert ist eine Funktion */
		FUNCTION
	}

	;

	private Double _doubleN;

	private Long _longN;

	private List<AttributeDescription> _attributes;

	private String _string;

	private Boolean _bool;

	/** Funktion */
	private ExpressionInterface _expressionObject;

	/** Operanden der Funktion (Jedes Listenelement entspricht einem Operanden) */
	private List<ExpressionTree> _expressionList;

	/** Typ der in diesem Blatt gespeichert wird. */
	private LeafType _type;

	/** Lege Blatt ohne Wert an. Operation wird auf invalid gesetzt. */
	private ExpressionLeaf() {
		super(null, null, Operation.invalid);
		_string = null;
		_doubleN = null;
		_attributes = null;
		_bool = null;
		_longN = null;
		_expressionObject = null;
		_expressionList = null;
	}

	/**
	 * Legt ein Blatt mit einem booleschen Wert an.
	 *
	 * @param b Wert des Blattes
	 */
	public ExpressionLeaf(final Boolean b) {
		this();
		_bool = b;
		_type = LeafType.BOOL;
	}

	/**
	 * Legt ein Blatt an, das einen String speichert
	 *
	 * @param s String
	 */
	public ExpressionLeaf(final String s) {
		this();
		_string = s;
		_type = LeafType.STRING;
	}

	/**
	 * Legt ein Blatt an, das ein Attribut speichert
	 *
	 * @param a Attribut
	 */
	public ExpressionLeaf(final AttributeDescription a) {
		this();
		_attributes = Collections.singletonList(a);
		_type = LeafType.ATTRIBUTE;
	}

	/**
	 * Legt ein Blatt an, das mehrere Attribute speichert. Dies ist der Fall, wenn auf ein Attribut verwiesen werden soll, dass z.B. Aliase benutzt und deshalb mehrere Objekte verwendet.
	 *
	 * @param a Attribut
	 */
	public ExpressionLeaf(final AttributeDescription[] a) {
		this();
		_attributes = Arrays.asList(a);
		_type = LeafType.ATTRIBUTE;
	}

	/**
	 * Legt ein Blatt an, das eine Fließkommazahl speichert
	 *
	 * @param d Double
	 */
	public ExpressionLeaf(final Double d) {
		this();
		_doubleN = d;
		_type = LeafType.DOUBLE;
	}

	/**
	 * Legt ein Blatt an, das eine Ganzzahl speichert
	 *
	 * @param n Ganzzahl
	 */
	public ExpressionLeaf(final Long n) {
		this();
		_longN = n;
		_type = LeafType.LONG;
	}

	/**
	 * Legt ein Blatt an, das eine Funktion beinhaltet
	 *
	 * @param functionName   Name der Klasse, die die Funktion repräsentiert
	 * @param expressionList Operanden der Funktion (Jedes Listenelement entspricht einem Operanden)
	 *
	 * @throws CriticalParserException Zur Funktion passende KLasse existiert nicht.
	 */
	public ExpressionLeaf(final String functionName, final List<ExpressionTree> expressionList) throws CriticalParserException {
		Class<?> function = null;
		Object obj = null;
		try {
			function = Class.forName(functionName);
			obj = function.newInstance();
		}
		catch(Exception e) {
			throw new CriticalParserException("Kann kein Objekt vom Typ " + functionName + " erzeugen " + e);
		}

		if(obj instanceof ExpressionInterface) {
			_expressionObject = (ExpressionInterface)obj;
		}
		else {
			throw new CriticalParserException(functionName + " ist keine Klasse vom Typ " + ExpressionInterface.class.getName());
		}

		this._expressionList = expressionList;
		_type = LeafType.FUNCTION;
	}

	/** @return Liste mit Attributen, die in diesem Blatt gespeichert werden. Falls kein Attribut abgelegt ist, wird leere Liste zurückgegeben */
	@Override
    public List<AttributeDescription> getAttributes() {
		final List<AttributeDescription> v = new ArrayList<AttributeDescription>();

		if(_attributes != null && _type == LeafType.ATTRIBUTE) {
			v.addAll(_attributes);
		}
		else if(_type == LeafType.FUNCTION) {
			for(final ExpressionTree expression : _expressionList) {
				v.addAll(expression.getAttributes());
			}
		}

		return v;
	}

	/** @see de.bsvrz.pua.prot.util.ExpressionTree#getAttributeOrder(ArrayList,HashSet) */
	@Override
	protected void getAttributeOrder(final ArrayList<AttributeDescription> list, final HashSet<AttributeDescription> visited) {
		if(_attributes != null && _type == LeafType.ATTRIBUTE) {
			if(!visited.contains(_attributes)) list.addAll(_attributes);
		}
		else if(_type == LeafType.FUNCTION) {
			for(final ExpressionTree expression : _expressionList) {
				list.addAll(expression.getAttributes());
			}
		}
	}

	/**
	 * Gibt den Ausdruck zurück, der in diesem Blatt abgelegt ist Ist es ein Attribut, wird der Wert aus der übergebenen Hashtable gesucht. Falls der Wert nicht
	 * bestimmt werden konnte, dann liefert {@link ExpressionResult#getType()} den Wert {@link de.bsvrz.pua.prot.util.ExpressionResult.ResultType#NONE}. Falls der
	 * Typ des Verwendeten realen Attributs nicht unterstützt wird (z.B. <code>ReferenceAttributeType</code>) ist result.type = error;
	 *
	 * @param realElements      Mapping der realen Elemente zu den ihnen zugeordeten Werten.
	 * @param tempElements      Mapping der temporären Attribute zu den ihnen zugeordneten Werten.
	 * @param pseudoObjectIndex Falls ein Attribut Aliase verwendet kann hier der Pseudoobjekt-Index festgelegt werden, von welchem Pseudoobjekt/Alias der Wert
	 *                          geholt werden soll. Ansonsten -1.
	 *
	 * @return Wert, der in diesem Blatt gespeichert ist.
	 */
	@Override
	public ExpressionResultAndState evaluateEx(
			final Hashtable<RealElement, BaseDataSet> realElements,
			final Hashtable<TempAttributeDescription, TempAttribut> tempElements,
			final int pseudoObjectIndex) {
		if(_type == LeafType.DOUBLE) {
			return ExpressionResultAndState.plain(_doubleN);
		}
		else if(_type == LeafType.LONG) {
			return ExpressionResultAndState.plain(_longN);
		}
		else if(_type == LeafType.BOOL) {
			return ExpressionResultAndState.plain(_bool);
		}
		else if(_type == LeafType.STRING) {
			return ExpressionResultAndState.plain(_string);
		}
		else if(_type == LeafType.FUNCTION) {
			ExpressionResult result;
			result = _expressionObject.evaluate(_expressionList, realElements, tempElements, pseudoObjectIndex);
			PuADataState state = PuADataState.COMPUTED;
			for(ExpressionTree expressionTree : _expressionList) {
				ExpressionResultAndState val = expressionTree.evaluateEx(realElements, tempElements, pseudoObjectIndex);
				if(val.getState().compareTo(state) > 0){
					state = val.getState();
				}
			}
			if(result.getType() == ExpressionResult.ResultType.ERROR){
				state = PuADataState.ERROR;
			}
			return new ExpressionResultAndState(result, state);
		}
		else if(_type == LeafType.ATTRIBUTE) {
			if(_attributes.get(0) instanceof RealAttributeDescription) {
				final int index;
				if(pseudoObjectIndex == -1) {
					// Einfacher Zugriff ohne PseudoObjekt-Index. Wenn dieses Attribut Aliase verwendet, Fehler zurückgeben.
					if(_attributes.size() > 1) {
						return ExpressionResultAndState.error(
								"Ungültiger Zugriff auf Temp-attribut mit Alias-Definition: " + this
						);
					}
					index = 0;
				}
				else {
					// Zugriff mit Index auf Alias-Definitionen
					index = pseudoObjectIndex;
				}
				final AttributeDescription attributeDescription = _attributes.get(index);
				final BaseDataSet value = realElements.get(attributeDescription);
				final ResultValue rv = value.getValue();
				if(rv.hasData()) {
					if(rv.isAtomar()) {
						return new ExpressionResultAndState(rv.getValue(), value.getDataState());
					}
					else {
						return ExpressionResultAndState.error(
								"Nicht-atomares Attribut im Ausdruck! : " + ((RealAttributeDescription)attributeDescription).getPrettyName()
						);
					}
				}
			}
			else if(_attributes.get(0) instanceof TempAttributeDescription) {
				return  ((TempAttributeDescription)_attributes.get(0)).getValue(realElements, tempElements, pseudoObjectIndex);
			}
			else if(_attributes.get(0) instanceof DurationAttributeDescription) {
				
				BaseDataSet baseDataSet = realElements.get(_attributes.get(0));
				return new ExpressionResultAndState(baseDataSet.getValue().getValue(), baseDataSet.getDataState());
			}
		}
		throw new IllegalStateException("Unbekannter Typ: " + _type);
	}

	/**
	 * Gibt zurück, welchen Datentyp dieses Blatt liefert. Wird während des Interpretierens des Skripts aufgerufen um Typkompatibilitäten sicherzustellen.
	 * @param ast AST
	 * @return Ergebnistyp
	 * @throws SemanticErrorException Falls ein Fehler auftritt (aufgrund eines ungültigen Skripts)
	 */
	@Override
	public ExpressionResult.ResultType getResultType(final AST ast) throws SemanticErrorException {
		switch(_type) {
			case STRING:
				return ExpressionResult.ResultType.STRING;
			case DOUBLE:
				return ExpressionResult.ResultType.DOUBLE;
			case LONG:
				return ExpressionResult.ResultType.LONG;
			case BOOL:
				return ExpressionResult.ResultType.BOOL;
			case FUNCTION:
				return _expressionObject.getResultType(_expressionList, ast);
			case ATTRIBUTE:
				final AttributeDescription attributeDescription = _attributes.get(0);
				if(attributeDescription instanceof RealAttributeDescription) {
					final RealAttributeDescription realAttributeDescription = (RealAttributeDescription)attributeDescription;
					final Attribute attribute = realAttributeDescription.getAttribute();
					return resultTypeFromAttributeType(attribute.getAttributeType(), ast);
				}
				if(attributeDescription instanceof TempAttributeDescription) {
					final TempAttributeDescription tempAttributeDescription = (TempAttributeDescription)attributeDescription;
					// Hier gibt es zwei Möglichkeiten: Entweder den Typ eines tempAttributes anhand des festgelegten Attributtyps bestimmen, oder die
					// Expression auswerten. Letzteres sorgt für weniger strenges Verhalten und wird deshalb derzeit bevorzugt.

//					final SystemObject type = tempAttributeDescription.type;
//					return resultTypeFromAttributeType(type, ast);

					// oder

					return tempAttributeDescription.getExpression().getResultType(ast);
				}
				if(attributeDescription instanceof DurationAttributeDescription) {
					return ExpressionResult.ResultType.LONG;
				}
				throw new SemanticErrorException(InterpreterErrorMessage.UNKNOWN_ATTRIBUTETYPE + attributeDescription.getName() + " (" + attributeDescription.getClass().getName() + ")", ast);
		}
		throw new SemanticErrorException(InterpreterErrorMessage.UNKNOWN_LEAFTYPE + _type, ast);
	}

	@Override
	public int getNumColumns() {
		if(_attributes != null && _type == LeafType.ATTRIBUTE) {
			final AttributeDescription attributeDescription = _attributes.get(0);
			if(attributeDescription instanceof RealAttributeDescription) {
				return _attributes.size();
			}
			if(attributeDescription instanceof TempAttributeDescription) {
				return ((TempAttributeDescription)attributeDescription).getExpression().getNumColumns();
			}
		}
		else if(_type == LeafType.FUNCTION) {
			int columns = 1;
			for(final ExpressionTree expression : _expressionList) {
				columns = Math.max(columns, expression.getNumColumns());
			}
			return columns;
		}
		return 1;
	}

	/**
	 * Gitbt anhand eines Attributtyps den erwarteteten Datentyp zurück. Wird während des Interpretierens des Skripts aufgerufen um Typkompatibilitäten sicherzustellen.
	 * @param type Attributtyp
	 * @param ast AST
	 * @return Datentyp
	 * @throws SemanticErrorException Falls ein Fehler auftritt
	 */
	private static ExpressionResult.ResultType resultTypeFromAttributeType(final SystemObject type, final AST ast) throws SemanticErrorException {
		if(type instanceof AttributeType) {
			final AttributeType attributeType = (AttributeType)type;
			if(attributeType instanceof StringAttributeType) {
				return ExpressionResult.ResultType.STRING;
			}
			if(attributeType instanceof IntegerAttributeType) {
				return ExpressionResult.ResultType.LONG;
			}
			if(attributeType instanceof DoubleAttributeType) {
				return ExpressionResult.ResultType.DOUBLE;
			}
			if(attributeType instanceof TimeAttributeType) {
				return ExpressionResult.ResultType.LONG;
			}
			if(attributeType instanceof ReferenceAttributeType) {
				return ExpressionResult.ResultType.LONG;
			}
			throw new SemanticErrorException(InterpreterErrorMessage.INVALID_TEMP_ATTRIBUTE_TYPE + attributeType, ast);
		}
		throw new SemanticErrorException(InterpreterErrorMessage.INVALID_ATTRIBUTETYPE_PID + type, ast);
	}

	/**
	 * Erzeugt ein Blatt das die ganzzahl 'number' speichert
	 *
	 * @param ast    Knoten, der die Zahl enthält (wird nur für die Zeilenangabe in der Fehlermeldung benötigt)
	 * @param number Ganz- oder Fliesskommazahl
	 *
	 * @return Blatt dass die Zahl enthält.
	 *
	 * @throws SemanticErrorException Falls 'number' eine ungültige Zahl ist
	 */
	static public ExpressionLeaf createLongLeaf(final AST ast, final String number) throws SemanticErrorException {
		ExpressionLeaf result = null;
		try {
			result = new ExpressionLeaf(new Long(number));
		}
		catch(NumberFormatException e) {
			throw new SemanticErrorException(number + ", " + e.getMessage(), ast);
		}
		return result;
	}

	/**
	 * Erzeugt ein Blatt das die Fließkommazahl 'number' speichert
	 *
	 * @param ast    Knoten, der die Zahl enthält (wird nur für die Zeilenangabe in der Fehlermeldung benötigt)
	 * @param number Ganz- oder Fliesskommazahl
	 *
	 * @return Blatt dass die Zahl enthält.
	 *
	 * @throws SemanticErrorException Falls 'number' eine ungültige Zahl ist
	 */
	static public ExpressionLeaf createDoubleLeaf(final AST ast, final String number) throws SemanticErrorException {
		ExpressionLeaf result = null;
		try {
			result = new ExpressionLeaf(new Double(number));
		}
		catch(NumberFormatException e) {
			throw new SemanticErrorException(InterpreterErrorMessage.NUMBER_EXPECTED + number + ", " + e.getMessage(), ast);
		}
		return result;
	}

	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		if(_type == LeafType.ATTRIBUTE) {
			return _attributes.get(0).getName();
		}
		else if(_type == LeafType.BOOL) {
			return _bool.toString();
		}
		else if(_type == LeafType.DOUBLE) {
			return _doubleN.toString();
		}
		else if(_type == LeafType.LONG) {
			return _longN.toString();
		}
		else if(_type == LeafType.STRING) {
			return _string;
		}
		else if(_type == LeafType.FUNCTION) {
			String result = "[" + _expressionObject.toString() + "(";
			if(_expressionList != null) {
				for(final ExpressionTree tree : _expressionList) {
					result += " " + tree.toString();
				}
			}
			return result + ")]";
		}
		else {
			return " error ";
		}
	}

	/** @see de.bsvrz.pua.prot.util.ExpressionTree#getFirstRealAttribute() */
	@Override
	public RealAttributeDescription getFirstRealAttribute() {
		if(_attributes != null && _attributes.get(0) instanceof RealAttributeDescription) return (RealAttributeDescription)_attributes.get(0);
		return null;
	}

	/** @see de.bsvrz.pua.prot.util.ExpressionTree#getFirstTempAttribute() */
	@Override
	public TempAttributeDescription getFirstTempAttribute() {
		if(_attributes != null && _attributes.get(0) instanceof TempAttributeDescription) return (TempAttributeDescription)_attributes.get(0);
		return null;
	}

	/** @see de.bsvrz.pua.prot.util.ExpressionTree#getDirectTempAttributes() */
	@Override
	public HashSet<TempAttributeDescription> getDirectTempAttributes() {
		final HashSet<TempAttributeDescription> result = new HashSet<TempAttributeDescription>();
		if(_attributes != null && _attributes.get(0) instanceof TempAttributeDescription) {
			for(final AttributeDescription attribute : _attributes) {
				result.add((TempAttributeDescription)attribute);
			}
		}

		return result;
	}

	/** @see de.bsvrz.pua.prot.util.ExpressionTree#getRealAttributes() */
	@Override
	public HashSet<RealAttributeDescription> getRealAttributes() {
		final HashSet<RealAttributeDescription> result = new HashSet<RealAttributeDescription>();
		if(_attributes != null) {
			for(final AttributeDescription attribute : _attributes) {
				if(attribute instanceof RealAttributeDescription) {
					result.add((RealAttributeDescription)attribute);
				}
				else if(attribute instanceof TempAttributeDescription) {
					TempAttributeDescription description = (TempAttributeDescription) attribute;
					result.addAll(description.getRealElements());
				}
			}
		}
		return result;
	}

	/** @see java.lang.Object#equals(java.lang.Object) */
	@Override
    public boolean equals(final Object other) {
		boolean result = false;
		if(other instanceof ExpressionLeaf) {
			final ExpressionLeaf leaf = (ExpressionLeaf)other;
			result = _type == leaf._type;

			if(_type == LeafType.STRING && _string != null) {
				result &= _string.equals(leaf._string);
			}
			else if(_type == LeafType.DOUBLE && _doubleN != null) {
				result &= _doubleN.equals(leaf._doubleN);
			}
			else if(_type == LeafType.LONG && _longN != null) {
				result &= _longN.equals(leaf._longN);
			}
			else if(_type == LeafType.ATTRIBUTE && _attributes != null) {
				result &= _attributes.equals(leaf._attributes);
			}
			else if(_type == LeafType.BOOL && _bool != null) {
				result &= _bool.equals(leaf._bool);
			}
			else if(_type == LeafType.FUNCTION && _expressionObject != null && _expressionList != null) {
				result &=
						_expressionObject.getClass().getName().equals(leaf._expressionObject.getClass().getName()) && _expressionList.equals(leaf._expressionList);
			}
		}
		return result;
	}

	/** @see de.bsvrz.pua.prot.util.ExpressionTree#containsAttribute() */
	@Override
	public boolean containsAttribute() {
		if(_type == LeafType.ATTRIBUTE) {
			return true;
		}
		else if(_type == LeafType.FUNCTION) {
			for(final ExpressionTree tree : _expressionList) {
				if(tree.containsAttribute()) return true;
			}
			return false;
		}
		else {
			return false;
		}
	}

	/** @return Liefert den Typ des Blattes. */
	public LeafType getType() {
		return _type;
	}

	/** @return Liefert den abgelegten boolschen Wert. */
	public Boolean getBool() {
		return _bool;
	}

	/** @return Liefert den abgelegten double Wert. */
	public Double getDouble() {
		return _doubleN;
	}

	/** @return Liefert den abgelegten Long Wert. */
	public Long getLong() {
		return _longN;
	}

	/** @return Liefert den abgelegten String. */
	public String getString() {
		return _string;
	}

	/**
	 * @return Liefert die Liste der Funktionsparameter.
	 *
	 * @see #getExpressionObject()
	 */
	public List<ExpressionTree> getExpressionList() {
		return _expressionList;
	}

	/**
	 * @return Liefert die abgelegte Funktion.
	 *
	 * @see #getExpressionList()
	 */
	public ExpressionInterface getExpressionObject() {
		return _expressionObject;
	}
}
